from typing import TYPE_CHECKING, List, Optional, Union
from pydantic import parse_obj_as
import logging
from urllib.parse import urljoin, urlsplit
import re

from .models import (
    VinDecodeResult, VinDecodeFlatResult, VinDecodeExtendedResult,
    VinDecodeExtendedFlatResult, WmiDecodeResult, WmiForManufacturerResult,
    AllMakesResult, PartsResult, AllManufacturersResult, ManufacturerDetailsResult,
    MakeForManufacturerResult, MakesForManufacturerAndYearResult, MakesForVehicleTypeResult,
    VehicleTypesForMakeResult, VehicleTypesForMakeIdResult, EquipmentPlantCodeResult,
    ModelsForMakeResult, ModelsForMakeIdResult, ModelsForMakeYearResult,
    VehicleVariableListResult, VehicleVariableValuesListResult, DecodeVinBatchResult,
    CanadianVehicleSpecificationsResult, BaseNHTSAResponse
)

if TYPE_CHECKING:
    from ...client import NhtsaClient

logger = logging.getLogger(__name__)


class VinDecodingAPI:
    """
    API for decoding Vehicle Identification Numbers (VINs) and related vehicle information.
    """
    def __init__(self, client: "NhtsaClient"):
        """
        Initializes the VinDecodingAPI.

        Args:
            client (NhtsaClient): The main client instance.
        """
        self.client = client

    async def decode_vin(self, vin: str, model_year: Optional[int] = None) -> VinDecodeResult:
        """
        Decodes the VIN and returns the output as Key-value pairs.

        Args:
            vin (str): The VIN to decode. Supports partial VINs (less than 17 characters).
            model_year (Optional[int]): The vehicle's model year (recommended for accuracy).

        Returns:
            VinDecodeResult: A Pydantic model representing the decoded VIN in key-value pairs.
        """
        params = {"format": "json"}
        if model_year:
            params["modelyear"] = model_year
        url = f"/vehicles/DecodeVin/{vin}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VinDecodeResult, response.json())

    async def decode_vin_flat_format(self, vin: str, model_year: Optional[int] = None) -> VinDecodeFlatResult:
        """
        Decodes the VIN and returns the output in a flat file format.

        Args:
            vin (str): The VIN to decode. Supports partial VINs.
            model_year (Optional[int]): The vehicle's model year (recommended for accuracy).

        Returns:
            VinDecodeFlatResult: A Pydantic model representing the decoded VIN in a flat format.
        """
        params = {"format": "json"}
        if model_year:
            params["modelyear"] = model_year
        url = f"/vehicles/DecodeVinValues/{vin}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VinDecodeFlatResult, response.json())

    async def decode_vin_extended(self, vin: str, model_year: Optional[int] = None) -> VinDecodeExtendedResult:
        """
        Decodes the VIN and provides additional information on variables related to other NHTSA programs.

        Args:
            vin (str): The VIN to decode. Supports partial VINs.
            model_year (Optional[int]): The vehicle's model year (recommended for accuracy).

        Returns:
            VinDecodeExtendedResult: A Pydantic model representing the decoded VIN with extended information.
        """
        params = {"format": "json"}
        if model_year:
            params["modelyear"] = model_year
        url = f"/vehicles/DecodeVinExtended/{vin}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VinDecodeExtendedResult, response.json())

    async def decode_vin_extended_flat_format(self, vin: str, model_year: Optional[int] = None) -> VinDecodeExtendedFlatResult:
        """
        Decodes the VIN with extended information and returns the output in a flat format.

        Args:
            vin (str): The VIN to decode. Supports partial VINs.
            model_year (Optional[int]): The vehicle's model year (recommended for accuracy).

        Returns:
            VinDecodeExtendedFlatResult: A Pydantic model representing the decoded VIN with extended information in a flat format.
        """
        params = {"format": "json"}
        if model_year:
            params["modelyear"] = model_year
        url = f"/vehicles/DecodeVinValuesExtended/{vin}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VinDecodeExtendedFlatResult, response.json())

    async def decode_wmi(self, wmi: str) -> WmiDecodeResult:
        """
        Provides information on the World Manufacturer Identifier (WMI) for a specific WMI code.

        Args:
            wmi (str): The WMI code (3 or 6 characters).

        Returns:
            WmiDecodeResult: A Pydantic model representing the decoded WMI information.
        """
        params = {"format": "json"}
        url = f"/vehicles/DecodeWMI/{wmi}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(WmiDecodeResult, response.json())

    async def get_wmis_for_manufacturer(self, manufacturer: str, vehicle_type: Optional[str] = None) -> WmiForManufacturerResult:
        """
        Provides information on all World Manufacturer Identifiers (WMI) for a specified Manufacturer.

        Args:
            manufacturer (str): Manufacturer's name (partial or full) or ID.
            vehicle_type (Optional[str]): Vehicle type name (partial or full) or ID to filter by.

        Returns:
            WmiForManufacturerResult: A Pydantic model representing a list of WMIs for the manufacturer.
        """
        params = {"format": "json"}
        if vehicle_type:
            params["vehicleType"] = vehicle_type
        url = f"/vehicles/GetWMIsForManufacturer/{manufacturer}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(WmiForManufacturerResult, response.json())

    async def get_all_makes(self) -> AllMakesResult:
        """
        Provides a list of all the Makes available in the vPIC Dataset.

        Returns:
            AllMakesResult: A Pydantic model representing a list of all makes.
        """
        params = {"format": "json"}
        url = "/vehicles/GetAllMakes"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(AllMakesResult, response.json())

    async def get_parts(self, type: int, from_date: Optional[str] = None, to_date: Optional[str] = None, manufacturer: Optional[str] = None, page: int = 1) -> PartsResult:
        """
        Provides a list of ORGs with letter date in the given range of the dates and with specified Type of ORG.

        Args:
            type (int): 565 (VIN Guidance) or 566 (Manufacturer Identification).
            from_date (Optional[str]): ORG's Letter Date should be on or after this date (MM/DD/YYYY).
            to_date (Optional[str]): ORG's Letter Date should be on or before this date (MM/DD/YYYY).
            manufacturer (Optional[str]): Manufacturer's name (partial or full) or ID.
            page (int): Page number for results (default 1, 1000 records per page).

        Returns:
            PartsResult: A Pydantic model representing a list of parts information.
        """
        params = {"type": type, "format": "json", "page": page}
        if from_date:
            params["fromDate"] = from_date
        if to_date:
            params["toDate"] = to_date
        if manufacturer:
            params["manufacturer"] = manufacturer
        url = "/vehicles/GetParts"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(PartsResult, response.json())

    async def get_all_manufacturers(self, manufacturer_type: Optional[Union[str, int]] = None, page: int = 1) -> AllManufacturersResult:
        """
        Returns a list of all the Manufacturers available in vPIC Dataset.

        Args:
            manufacturer_type (Optional[Union[str, int]]): Manufacturer Type to filter by (e.g., "Intermediate" or ID).
            page (int): Page number for results (default 1, 100 items per page).

        Returns:
            AllManufacturersResult: A Pydantic model representing a list of manufacturers.
        """
        params = {"format": "json", "page": page}
        if manufacturer_type:
            params["ManufacturerType"] = manufacturer_type
        url = "/vehicles/GetAllManufacturers"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(AllManufacturersResult, response.json())

    async def get_manufacturer_details(self, manufacturer: Union[str, int], page: int = 1) -> ManufacturerDetailsResult:
        """
        Provides the details for a specific manufacturer(s).

        Args:
            manufacturer (Union[str, int]): Manufacturer's name (partial or full) or ID.
            page (int): Page number for results (default 1, 100 items per page if manufacturer is a string).

        Returns:
            ManufacturerDetailsResult: A Pydantic model representing the manufacturer details.
        """
        params = {"format": "json"}
        if isinstance(manufacturer, str) and not manufacturer.isdigit():
            params["page"] = page
        url = f"/vehicles/GetManufacturerDetails/{manufacturer}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(ManufacturerDetailsResult, response.json())

    async def get_makes_for_manufacturer(self, manufacturer: Union[str, int]) -> MakeForManufacturerResult:
        """
        Returns all the Makes in the vPIC dataset for a specified manufacturer.

        Args:
            manufacturer (Union[str, int]): Manufacturer's name (partial or full) or ID.

        Returns:
            MakeForManufacturerResult: A Pydantic model representing a list of makes for the manufacturer.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetMakeForManufacturer/{manufacturer}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(MakeForManufacturerResult, response.json())

    async def get_makes_for_manufacturer_and_year(self, manufacturer: Union[str, int], year: int) -> MakesForManufacturerAndYearResult:
        """
        Returns all the Makes in the vPIC dataset for a specified manufacturer and
        whose Year From and Year To range cover the specified year.

        Args:
            manufacturer (Union[str, int]): Manufacturer's name (partial or full) or ID.
            year (int): The year to filter by.

        Returns:
            MakesForManufacturerAndYearResult: A Pydantic model representing a list of makes for the manufacturer and year.
        """
        params = {"format": "json", "year": year}
        url = f"/vehicles/GetMakesForManufacturerAndYear/{manufacturer}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(MakesForManufacturerAndYearResult, response.json())

    async def get_makes_for_vehicle_type(self, vehicle_type_name: str) -> MakesForVehicleTypeResult:
        """
        Returns all the Makes in the vPIC dataset for a specified vehicle type whose name is LIKE the vehicle type name.

        Args:
            vehicle_type_name (str): Vehicle type name (partial or full).

        Returns:
            MakesForVehicleTypeResult: A Pydantic model representing a list of makes for the vehicle type.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetMakesForVehicleType/{vehicle_type_name}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(MakesForVehicleTypeResult, response.json())

    async def get_vehicle_types_for_make(self, make_name: str) -> VehicleTypesForMakeResult:
        """
        Returns all the Vehicle Types in the vPIC dataset for a specified Make whose name is LIKE the make name.

        Args:
            make_name (str): Make name (partial or full).

        Returns:
            VehicleTypesForMakeResult: A Pydantic model representing a list of vehicle types for the make.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetVehicleTypesForMake/{make_name}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VehicleTypesForMakeResult, response.json())

    async def get_vehicle_types_for_make_id(self, make_id: int) -> VehicleTypesForMakeIdResult:
        """
        Returns all the Vehicle Types in the vPIC dataset for a specified Make whose ID equals the make ID.

        Args:
            make_id (int): The ID of the make.

        Returns:
            VehicleTypesForMakeIdResult: A Pydantic model representing a list of vehicle types for the make ID.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetVehicleTypesForMakeId/{make_id}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VehicleTypesForMakeIdResult, response.json())

    async def get_equipment_plant_codes(self, year: int, equipment_type: Optional[int] = None, report_type: Optional[str] = None) -> EquipmentPlantCodeResult:
        """
        Returns assigned Equipment Plant Codes. Can be filtered by Year, Equipment Type and Report Type.

        Args:
            year (int): Only years 2016 and above are supported.
            equipment_type (Optional[int]): Equipment type (e.g., 1 for Tires).
            report_type (Optional[str]): Report type (e.g., "New", "Updated", "Closed", "All").

        Returns:
            EquipmentPlantCodeResult: A Pydantic model representing a list of equipment plant codes.
        """
        params = {"format": "json", "year": year}
        if equipment_type:
            params["equipmentType"] = equipment_type
        if report_type:
            params["reportType"] = report_type
        url = "/vehicles/GetEquipmentPlantCodes"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(EquipmentPlantCodeResult, response.json())

    async def get_models_for_make(self, make_name: str) -> ModelsForMakeResult:
        """
        Returns the Models in the vPIC dataset for a specified Make whose name is LIKE the Make.

        Args:
            make_name (str): Make name (partial or full, or "*" for all makes).

        Returns:
            ModelsForMakeResult: A Pydantic model representing a list of models for the make.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetModelsForMake/{make_name}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(ModelsForMakeResult, response.json())

    async def get_models_for_make_id(self, make_id: int) -> ModelsForMakeIdResult:
        """
        Returns the Models in the vPIC dataset for a specified Make whose Id is EQUAL the MakeId.

        Args:
            make_id (int): The ID of the make (or 0 for all makes).

        Returns:
            ModelsForMakeIdResult: A Pydantic model representing a list of models for the make ID.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetModelsForMakeId/{make_id}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(ModelsForMakeIdResult, response.json())

    async def get_models_for_make_year(self, make_name: str, model_year: Optional[int] = None, vehicle_type: Optional[str] = None) -> ModelsForMakeYearResult:
        """
        Returns the Models in the vPIC dataset for a specified year and Make.

        Args:
            make_name (str): Make name (partial or full).
            model_year (Optional[int]): Model year (greater than 1995).
            vehicle_type (Optional[str]): Vehicle type name (partial or full).

        Returns:
            ModelsForMakeYearResult: A Pydantic model representing a list of models.
        """
        params = {"format": "json"}
        url_segments = []
        if make_name:
            url_segments.append(f"make/{make_name}")
        if model_year:
            url_segments.append(f"modelyear/{model_year}")
        if vehicle_type:
            url_segments.append(f"vehicletype/{vehicle_type}")

        url = f"/vehicles/GetModelsForMakeYear/{'/'.join(url_segments)}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(ModelsForMakeYearResult, response.json())

    async def get_models_for_make_id_year(self, make_id: int, model_year: Optional[int] = None, vehicle_type: Optional[str] = None) -> ModelsForMakeYearResult:
        """
        Returns the Models in the vPIC dataset for a specified year and Make ID.

        Args:
            make_id (int): The ID of the make.
            model_year (Optional[int]): Model year (greater than 1995).
            vehicle_type (Optional[str]): Vehicle type name (partial or full).

        Returns:
            ModelsForMakeYearResult: A Pydantic model representing a list of models.
        """
        params = {"format": "json"}
        url_segments = []
        url_segments.append(f"makeId/{make_id}")
        if model_year:
            url_segments.append(f"modelyear/{model_year}")
        if vehicle_type:
            url_segments.append(f"vehicletype/{vehicle_type}")

        url = f"/vehicles/GetModelsForMakeIdYear/{'/'.join(url_segments)}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(ModelsForMakeYearResult, response.json())

    async def get_vehicle_variable_list(self) -> VehicleVariableListResult:
        """
        Provides a list of all the Vehicle related variables that are in vPIC dataset.

        Returns:
            VehicleVariableListResult: A Pydantic model representing a list of vehicle variables.
        """
        params = {"format": "json"}
        url = "/vehicles/GetVehicleVariableList"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VehicleVariableListResult, response.json())

    async def get_vehicle_variable_values_list(self, variable_search_param: Union[str, int]) -> VehicleVariableValuesListResult:
        """
        Provides a list of all the accepted values for a given variable that are stored in vPIC dataset.

        Args:
            variable_search_param (Union[str, int]): Variable Name (full name) or Variable ID.

        Returns:
            VehicleVariableValuesListResult: A Pydantic model representing a list of variable values.
        """
        params = {"format": "json"}
        url = f"/vehicles/GetVehicleVariableValuesList/{variable_search_param}"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(VehicleVariableValuesListResult, response.json())

    async def decode_vin_batch(self, data: str) -> DecodeVinBatchResult:
        """
        Decodes a batch of VINs that are submitted in a standardized format in a string to return multiple decodes.

        Args:
            data (str): The input string in the format "vin,modelYear;vin,modelYear;...".

        Returns:
            DecodeVinBatchResult: A Pydantic model representing the decoded VINs in a batch.
        """
        post_fields = {'format': 'json', 'data': data}
        url = "/vehicles/DecodeVINValuesBatch/"
        response = await self.client._request("POST", url, data=post_fields, use_vpic_client=True)
        return parse_obj_as(DecodeVinBatchResult, response.json())

    async def get_canadian_vehicle_specifications(self, year: int, make: str, model: Optional[str] = None, units: Optional[str] = None) -> CanadianVehicleSpecificationsResult:
        """
        The Canadian Vehicle Specifications (CVS) consists of a database of original vehicle dimensions.

        Args:
            year (int): Number, >= 1971.
            make (str): Vehicle's make, like "Honda", "Toyota".
            model (Optional[str]): Vehicle's model, like "Pilot", "Focus".
            units (Optional[str]): "Metric" (default), or "US" for United States customary units.

        Returns:
            CanadianVehicleSpecificationsResult: A Pydantic model representing Canadian vehicle specifications.
        """
        params = {"format": "json", "Year": year, "Make": make}
        if model:
            params["Model"] = model
        if units:
            params["units"] = units
        url = "/vehicles/GetCanadianVehicleSpecifications/"
        response = await self.client._request("GET", url, params=params, use_vpic_client=True)
        return parse_obj_as(CanadianVehicleSpecificationsResult, response.json())

    async def get_standalone_vpic_db_url(self) -> str:
        """
        Scrapes the vPIC API homepage to find the standalone VIN decoding database URL (.bak.zip).

        Args:
            None

        Returns:
            str: Absolute URL to the current .bak.zip file. Falls back to the current raw URL if not found.

        Raises:
            None

        Examples:
            >>> url = await client.vin_decoding.get_standalone_vpic_db_url()
        """
        default_url = f"{self.client.VPIC_BASE_URL}/vPICList_lite_2025_09.bak.zip"
        try:
            response = await self.client._request(
                "GET",
                "/",
                use_vpic_client=True,
                headers={"accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"},
            )
            html = response.text
            match = re.search(r'href="(?P<href>/api/[^"]+\.bak\.zip)"', html, re.IGNORECASE)
            if match:
                href = match.group("href")
                parts = urlsplit(self.client.VPIC_BASE_URL)
                origin = f"{parts.scheme}://{parts.netloc}"
                return urljoin(origin + "/", href.lstrip("/"))
            logger.warning("Standalone vPIC DB URL not found on index page; using default fallback.")
            return default_url
        except Exception as e:
            logger.error(f"Error scraping vPIC index for standalone DB URL: {e}", exc_info=True)
            return default_url



from typing import List, Optional
from pydantic import BaseModel, Field
from datetime import datetime


class InvestigationFlatFile(BaseModel):
    """
    Represents information about an available defect investigation flat file for download.
    """
    path: str = Field(..., description="The URL path to the investigation flat file.")
    size: str = Field(..., description="The size of the flat file.")
    updated: str = Field(..., description="The last updated date of the flat file.")

# Since the Investigations API primarily deals with downloading flat files as per the provided context,
# there are no direct JSON API responses to model beyond the metadata of the files themselves.
# If there were direct API endpoints for querying investigations (like for recalls or complaints),
# additional Pydantic models would be defined here to represent those response structures.


from typing import TYPE_CHECKING, List, Optional, Union
from pydantic import parse_obj_as
import logging
from datetime import date, datetime

from .models import (
    VehicleByYmmtResult, VehicleDetailsResult, VehicleByVinProxyResult,
    VehicleSearchResult, TireSearchResult, EquipmentSearchResult,
    ChildSeatsResult, ChildSeatModesResult, ChildSeatBySearchResult
)

if TYPE_CHECKING:
    from ...client import NhtsaClient

logger = logging.getLogger(__name__)


class ProductsAPI:
    """
    API for querying various NHTSA product information, including vehicles, tires,
    equipment, and child seats. This consolidates data from multiple product-focused
    endpoints within the main NHTSA API.
    """
    def __init__(self, client: "NhtsaClient"):
        """
        Initializes the ProductsAPI.

        Args:
            client (NhtsaClient): The main client instance.
        """
        self.client = client

    # --- Vehicle Endpoints ---

    async def get_vehicle_by_ymmt(
        self,
        model_year: int,
        make: str,
        model: str,
        trim: Optional[str] = None,
        series: Optional[str] = None,
        name: Optional[str] = None, # Unclear what 'name' refers to in byYmmt, often empty
        data: Optional[str] = "crashtestratings,safetyfeatures,recommendedfeatures", # Comma-separated list of data to include, unclear all options
        product_detail: Optional[str] = "all" # 'minimal' or 'all'
    ) -> VehicleByYmmtResult:
        """
        Retrieves vehicle information and safety details by Year, Make, Model, Trim.
        Discovered here: https://www.nhtsa.gov/vehicle

        Args:
            model_year (int): The model year of the vehicle.
            make (str): The make of the vehicle.
            model (str): The model of the vehicle.
            trim (Optional[str]): The trim level of the vehicle.
            series (Optional[str]): The series of the vehicle.
            name (Optional[str]): Additional name component.
            data (Optional[str]): Comma-separated list of data to include (e.g., "crashtestratings,safetyfeatures,recommendedfeatures").
            product_detail (Optional[str]): Level of product detail ("minimal" or "all").

        Returns:
            VehicleByYmmtResult: A Pydantic model representing the vehicle details.
        """
        logger.warning("If you are using this api directly from a vin decoder it is unlikely to work because NHTSA does not correctly map their vin decoder to the ymmt values used for this api.")
        url = f"/vehicles/byYmmt"
        params = {
            "modelYear": model_year,
            "make": make,
            "model": model,
            "data": data,
            "productDetail": product_detail
        }
        if trim:
            params["trim"] = trim
        if series:
            params["series"] = series
        if name: # The example showed 'name=' empty, so it might be optional and sometimes passed.
            params["name"] = name
        # logger.info("Full URL being requested: https://api.nhtsa.gov" + url + "?" + "&".join([f"{k}={v}" for k,v in params.items()]))

        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(VehicleByYmmtResult, response.json())

    async def get_vehicle_details_by_id(
        self,
        vehicle_id: int,
        data: Optional[str] = "complaints,recalls,investigations,manufacturercommunications",
        product_detail: Optional[str] = "minimal", # 'minimal' or 'all'
        name: Optional[str] = None # Unclear what 'name' refers to here.
    ) -> VehicleDetailsResult:
        """
        Retrieves detailed information (including safety issues) for a specific vehicle ID.

        Args:
            vehicle_id (int): The unique ID of the vehicle.
            data (Optional[str]): Comma-separated list of safety issues to include.
            product_detail (Optional[str]): Level of product detail ("minimal" or "all").
            name (Optional[str]): Additional name component.

        Returns:
            VehicleDetailsResult: A Pydantic model representing the vehicle's safety issue details.
        """
        url = f"/vehicles/{vehicle_id}/details"
        params = {
            "data": data,
            "productDetail": product_detail
        }
        if name:
            params["name"] = name
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(VehicleDetailsResult, response.json())

    async def get_vehicle_by_vin_proxy(
        self,
        vin: str,
        timestamp: int,
        key: int,
        # token: str, # This requires a captcha token, handled by raising NotImplementedError
        product_detail: Optional[str] = "all",
        data: Optional[str] = "none" # Default to 'none' as per example
    ) -> VehicleByVinProxyResult:
        """
        Retrieves vehicle information by VIN using a proxy endpoint that typically
        requires a captcha token. This method will raise NotImplementedError.
        API comes from here: https://www.nhtsa.gov/recalls

        Args:
            vin (str): The VIN to look up.
            timestamp (int): A timestamp required by the API.
            key (int): A key required by the API.
            # token (str): The captcha token.
            product_detail (Optional[str]): Level of product detail ("minimal" or "all").
            data (Optional[str]): Comma-separated list of data to include.

        Returns:
            VehicleByVinProxyResult: A Pydantic model representing the vehicle information.

        Raises:
            NotImplementedError: This endpoint currently requires a captcha token
                                 which is not handled by the SDK.
        """
        logger.warning("This api has a tendency to fail for some VINs, to no fault of our own. NHTSA vin decoding system does not correctly map the vin decoding --> vehicle id (which is cross database for them) for select vins. See ")
        
        # This api internally uses the vpic decoder (and the vehicle id by ymmt apis), but the vpic decoder does not match the ymmt values used (they must be cross database) to get the vehicle id.
        # For the same reason the tool call we made to replicate this `get_tsb_by_mfg_number` fails (which also uses the vin --> ymmt --> vehicle id workflow).
        # https://api.nhtsa.gov/vehicles/byVinProxy?productDetail=all&data=none&vin=1GCUDGE89PZ182915&timestamp=1760321417481&key=-547524275
        # {
        #     "meta": {
        #         "decoder": [
        #             {
        #                 "errors": [],
        #                 "make": "CHEVROLET",
        #                 "modelYear": "2023",
        #                 "source": "vPIC",
        #                 "sourceUrl": "https://vpic.nhtsa.dot.gov/api/vehicles/decodevinvalues/1GCUDGE89PZ182915?format=json",
        #                 "vehicleModel": "Silverado",
        #                 "vehicleType": "TRUCK"
        #             }
        #         ],
        #         "filters": null,
        #         "isWebApi": false,
        #         "messages": [
        #             {
        #                 "errorDetails": null,
        #                 "errorMessage": "There are no NCAP or Artemis vehicles in the unified database for the supplied VIN: 1GCUDGE89PZ182915, details: [[DriveType:4WD/4-Wheel Drive/4x4, Make:CHEVROLET, Model:Silverado, ModelYear:2023, VIN:1GCUDGE89PZ182915]], productStatus: ",
        #                 "errorName": "Request was invalid",
        #                 "specificHttpErrorCode": 400
        #             }
        #         ],
        #         "pagination": {
        #             "count": 0,
        #             "currentUrl": "https://api.nhtsa.gov/vehicles/byVin?offset=0&max=20&sort=id&productDetail=all&data=none&vin*****************",
        #             "grailsServerUrl": "https://api.nhtsa.gov",
        #             "max": 20,
        #             "nextUrl": null,
        #             "offset": 0,
        #             "order": null,
        #             "previousUrl": null,
        #             "sort": "id",
        #             "total": 0
        #         },
        #         "status": 400
        #     },
        #     "results": []
        # }

        # Example of working request:
        # https://api.nhtsa.gov/vehicles/byVinProxy?productDetail=all&data=none&vin=3GNMCFE08CG271867&timestamp=1760321531010&key=-8912962
        # {
        #     "meta": {
        #         "status": 200,
        #         "messages": [],
        #         "pagination": {
        #             "count": 1,
        #             "max": 20,
        #             "offset": 0,
        #             "sort": "id",
        #             "order": null,
        #             "total": 1,
        #             "currentUrl": "https://api.nhtsa.gov/vehicles/byVin?offset=0&max=20&sort=id&productDetail=all&data=none&vin*****************",
        #             "nextUrl": null,
        #             "previousUrl": null
        #         },
        #         "filters": null,
        #         "decoder": [
        #             {
        #                 "modelYear": "2012",
        #                 "make": "CHEVROLET",
        #                 "errors": [],
        #                 "sourceUrl": "https://vpic.nhtsa.dot.gov/api/vehicles/decodevinvalues/3GNMCFE08CG271867?format=json",
        #                 "vehicleModel": "Avalanche",
        #                 "vehicleType": "MULTIPURPOSE PASSENGER VEHICLE (MPV)",
        #                 "source": "vPIC"
        #             }
        #         ]
        #     },
        #     "results": [
        #         {
        #             "vehicleId": 6259, // As you can see this properly gets the vehicle id
        #             "parkIt": false,
        #             "parkOutSide": false,
        #             "overTheAirUpdate": false,
        #             "artemisId": 1450782,
        #             "active": false,
        #             "ncapId": 6452,
        #             "modelYear": 2012,
        #             "make": "CHEVROLET",
        #             "vehicleModel": "AVALANCHE",
        #             "trim": "SUV",
        #             "series": "RWD",
        #             "class": "SUV",
        #             "manufacturer": "GENERAL MOTORS CORP.",
        #             "vehiclePicture": "https://static.nhtsa.gov/images/vehicles/7729_st0640_046.png",
        #             "ncapRated": true,
        #             "complaintsCount": 52,
        #             "recallsCount": 3,
        #             "investigationsCount": 1,
        #             "manufacturerCommunicationsCount": 1603
        #         }
        #     ]
        # }
        
        
        # Construct URL for demonstration/error raising
        url = f"/vehicles/byVinProxy"
        params = {
            "vin": vin,
            "timestamp": timestamp,
            "key": key,
            "productDetail": product_detail,
            "data": data
        }
        # if token:
        #     params["token"] = token # Add token if available, but we'll raise anyway
        
        # This endpoint uses a CAPTCHA. Raising an error as per instructions.
        # In a real-world scenario, you would integrate a CAPTCHA solver here.
        raise NotImplementedError(
            "The 'byVinProxy' endpoint requires a Google reCAPTCHA token "
            "which is not currently handled by this SDK. "
            "URL built: https://api.nhtsa.gov" + url + "?" + "&".join([f"{k}={v}" for k,v in params.items()])
        )
        # response = await self.client._request("GET", url, params=params)
        # return parse_obj_as(VehicleByVinProxyResult, response.json())

    async def search_vehicles(
        self,
        query: str,
        offset: int = 0,
        max_results: int = 10,
        sort_by: Optional[str] = "overallRating",
        order: Optional[str] = "desc", # 'asc' or 'desc'
        data: Optional[str] = "crashtestratings,recommendedfeatures,safetyfeatures",
        date_start: Optional[Union[str, date, datetime]] = None,
        date_end: Optional[Union[str, date, datetime]] = None,
        product_detail: Optional[str] = "all", # 'minimal' or 'all'
        vehicle_class: Optional[str] = None, # Filter from 'filters' in response, e.g., 'TRUCK'
    ) -> VehicleSearchResult:
        """
        Searches for vehicles based on a query string and various filters, including safety data.

        Args:
            query (str): The search query (e.g., "chevrolet colorado").
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 10).
            sort_by (Optional[str]): Field to sort the results by (e.g., "overallRating", "make").
            order (Optional[str]): Sort order ("asc" or "desc").
            data (Optional[str]): Comma-separated list of data to include (e.g., "crashtestratings,safetyfeatures,recommendedfeatures").
            date_start (Optional[Union[str, date, datetime]]): Start date for filtering (YYYY-MM-DD).
            date_end (Optional[Union[str, date, datetime]]): End date for filtering (YYYY-MM-DD).
            product_detail (Optional[str]): Level of product detail ("minimal" or "all").
            vehicle_class (Optional[str]): Filter by vehicle class (e.g., "TRUCK", "PASSENGER CAR").

        Returns:
            VehicleSearchResult: A Pydantic model representing the search results.
        """
        url = f"/vehicles/bySearch"
        params = {
            "query": query,
            "offset": offset,
            "max": max_results,
            "sort": sort_by,
            "order": order,
            "data": data,
            "productDetail": product_detail
        }
        if date_start:
            if isinstance(date_start, date) or isinstance(date_start, datetime):
                params["dateStart"] = date_start.isoformat()
            else:
                params["dateStart"] = date_start
        if date_end:
            if isinstance(date_end, date) or isinstance(date_end, datetime):
                params["dateEnd"] = date_end.isoformat()
            else:
                params["dateEnd"] = date_end
        if vehicle_class:
            params["vehicleClass"] = vehicle_class
        
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(VehicleSearchResult, response.json())


    # --- Tire Endpoints ---

    async def search_tires(
        self,
        query: str,
        data_set: Optional[str] = "safetyIssues", # "ratings" or "safetyIssues"
        offset: int = 0,
        max_results: int = 10,
        sort_by: Optional[str] = "productName",
        order: Optional[str] = "asc", # 'asc' or 'desc'
        data: Optional[str] = "none" # "modes" etc.
    ) -> TireSearchResult:
        """
        Searches for tires based on a query string.

        Args:
            query (str): The search query (e.g., "goodyear wrangler").
            data_set (Optional[str]): The dataset to query ("ratings" or "safetyIssues").
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 10).
            sort_by (Optional[str]): Field to sort the results by (e.g., "productName").
            order (Optional[str]): Sort order ("asc" or "desc").
            data (Optional[str]): Additional data to include (e.g., "modes" for child seats, "none" for tires if not specified).

        Returns:
            TireSearchResult: A Pydantic model representing the tire search results.
        """
        url = f"/tires/bySearch"
        params = {
            "query": query,
            "dataSet": data_set,
            "offset": offset,
            "max": max_results,
            "sort": sort_by,
            "order": order,
            "data": data
        }
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(TireSearchResult, response.json())

    # --- Equipment Endpoints ---

    async def search_equipment(
        self,
        query: str,
        offset: int = 0,
        max_results: int = 10,
        sort_by: Optional[str] = "productName",
        order: Optional[str] = "asc", # 'asc' or 'desc'
        data: Optional[str] = "all" # e.g. "all" or specific fields
    ) -> EquipmentSearchResult:
        """
        Searches for equipment based on a query string.

        Args:
            query (str): The search query (e.g., "seatbelts").
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 10).
            sort_by (Optional[str]): Field to sort the results by (e.g., "productName").
            order (Optional[str]): Sort order ("asc" or "desc").
            data (Optional[str]): Additional data to include (e.g., "all").

        Returns:
            EquipmentSearchResult: A Pydantic model representing the equipment search results.
        """
        url = f"/equipment/bySearch"
        params = {
            "query": query,
            "offset": offset,
            "max": max_results,
            "sort": sort_by,
            "order": order,
            "data": data
        }
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(EquipmentSearchResult, response.json())


    # --- Child Seat Endpoints ---

    async def get_child_seats(
        self,
        data: Optional[str] = "modes", # e.g., "modes"
        data_set: Optional[str] = "ratings", # e.g., "ratings"
        weight: Optional[int] = None,
        height: Optional[int] = None,
        offset: int = 0,
        max_results: int = 10,
        sort_by: Optional[str] = "make",
        order: Optional[str] = "asc", # 'asc' or 'desc'
    ) -> ChildSeatsResult:
        """
        Retrieves child seat information, optionally filtered by weight and height.

        Args:
            data (Optional[str]): Data to include (e.g., "modes").
            data_set (Optional[str]): Dataset to query (e.g., "ratings").
            weight (Optional[int]): Child's weight in pounds.
            height (Optional[int]): Child's height in inches.
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 10).
            sort_by (Optional[str]): Field to sort the results by (e.g., "make").
            order (Optional[str]): Sort order ("asc" or "desc").

        Returns:
            ChildSeatsResult: A Pydantic model representing the child seat information.
        """
        url = f"/childSeats"
        params = {
            "offset": offset,
            "max": max_results,
            "sort": sort_by,
            "order": order,
            "data": data,
            "dataSet": data_set
        }
        if weight is not None:
            params["weight"] = weight
        if height is not None:
            params["height"] = height
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(ChildSeatsResult, response.json())

    async def get_child_seat_modes(
        self,
        offset: int = 0,
        max_results: int = 20,
        sort_by: Optional[str] = "id",
        order: Optional[str] = None # 'asc' or 'desc'
    ) -> ChildSeatModesResult:
        """
        Retrieves a list of available child seat modes.

        Args:
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 20).
            sort_by (Optional[str]): Field to sort the results by (e.g., "id").
            order (Optional[str]): Sort order ("asc" or "desc").

        Returns:
            ChildSeatModesResult: A Pydantic model representing the list of child seat modes.
        """
        url = f"/childSeats/modes"
        params = {
            "offset": offset,
            "max": max_results,
            "sort": sort_by
        }
        if order:
            params["order"] = order
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(ChildSeatModesResult, response.json())

    async def search_child_seats(
        self,
        query: str,
        data: Optional[str] = "modes", # e.g., "modes"
        data_set: Optional[str] = "ratings", # e.g., "ratings"
        offset: int = 0,
        max_results: int = 10,
        sort_by: Optional[str] = "make",
        order: Optional[str] = "asc", # 'asc' or 'desc'
    ) -> ChildSeatBySearchResult:
        """
        Searches for child seats based on a query string.

        Args:
            query (str): The search query (e.g., "chicco").
            data (Optional[str]): Data to include (e.g., "modes").
            data_set (Optional[str]): Dataset to query (e.g., "ratings").
            offset (int): The starting offset for results (default 0).
            max_results (int): The maximum number of results to return (default 10).
            sort_by (Optional[str]): Field to sort the results by (e.g., "make").
            order (Optional[str]): Sort order ("asc" or "desc").

        Returns:
            ChildSeatBySearchResult: A Pydantic model representing the child seat search results.
        """
        url = f"/childSeats/bySearch"
        params = {
            "query": query,
            "offset": offset,
            "max": max_results,
            "sort": sort_by,
            "order": order,
            "data": data,
            "dataSet": data_set
        }
        response = await self.client._request("GET", url, params=params)
        return parse_obj_as(ChildSeatBySearchResult, response.json())



from typing import List, Optional, Any, Dict, Union
from pydantic import BaseModel, Field
from datetime import datetime

from ...lib.models import Meta, Pagination, Error # Reusing common meta/pagination

# --- Common nested models for safety issues (Complaints, Recalls, Investigations, Mfr Comms) ---
class SafetyIssueComponent(BaseModel):
    id: int = Field(alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")

class SafetyIssueAssociatedDocument(BaseModel):
    id: Optional[int] = Field(alias="id", default=None)
    file_name: Optional[str] = Field(alias="fileName", default=None)
    file_size: Optional[int] = Field(alias="fileSize", default=None)
    load_date: Optional[datetime] = Field(alias="loadDate", default=None)
    meta_data: Optional[List[Dict[str, Any]]] = Field(alias="metaData", default_factory=list)
    mime_type: Optional[str] = Field(alias="mimeType", default=None)
    summary: Optional[str] = Field(alias="summary", default=None)
    url: Optional[str] = Field(alias="url", default=None)

class SafetyIssueAssociatedProduct(BaseModel):
    type: Optional[str] = Field(alias="type", default=None)
    product_year: Optional[str] = Field(alias="productYear", default=None)
    product_make: Optional[str] = Field(alias="productMake", default=None)
    product_model: Optional[str] = Field(alias="productModel", default=None)
    manufacturer: Optional[str] = Field(alias="manufacturer", default=None)
    size: Optional[str] = Field(alias="size", default=None) # Specific to Tires, but optional

class SafetyIssueComplaint(BaseModel):
    date_filed: Optional[datetime] = Field(alias="dateFiled", default=None)
    date_of_incident: Optional[datetime] = Field(alias="dateOfIncident", default=None)
    nhtsa_id_number: int = Field(alias="nhtsaIdNumber")
    id: int = Field(alias="id")
    number_of_injuries: int = Field(alias="numberOfInjuries")
    number_of_deaths: int = Field(alias="numberOfDeaths")
    fire: bool = Field(alias="fire")
    crash: bool = Field(alias="crash")
    vin: Optional[str] = Field(alias="vin", default=None)
    consumer_location: Optional[str] = Field(alias="consumerLocation", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    components: List[SafetyIssueComponent] = Field(alias="components", default_factory=list)
    associated_documents_count: int = Field(alias="associatedDocumentsCount")
    associated_documents: Union[List[SafetyIssueAssociatedDocument], str] = Field(alias="associatedDocuments", default_factory=list) # Can be list or string "N/A"
    associated_products_count: int = Field(alias="associatedProductsCount")
    associated_products: Union[List[SafetyIssueAssociatedProduct], str] = Field(alias="associatedProducts", default_factory=list) # Can be list or string "N/A"

class SafetyIssueRecallInvestigationInfo(BaseModel):
    nhtsa_campaign_number: Optional[str] = Field(alias="nhtsaCampaignNumber", default=None)
    subject: Optional[str] = Field(alias="subject", default=None)
    report_received_date: Optional[datetime] = Field(alias="reportReceivedDate", default=None)
    summary: Optional[str] = Field(alias="summary", default=None)

class SafetyIssueRecall(BaseModel):
    id: int = Field(alias="id")
    park_it: bool = Field(alias="parkIt")
    park_outside: bool = Field(alias="parkOutSide")
    over_the_air_update: bool = Field(alias="overTheAirUpdate")
    manufacturer: Optional[str] = Field(alias="manufacturer", default=None)
    mfr_campaign_number: Optional[str] = Field(alias="mfrCampaignNumber", default=None)
    nhtsa_campaign_number: Optional[str] = Field(alias="nhtsaCampaignNumber", default=None)
    report_received_date: Optional[datetime] = Field(alias="reportReceivedDate", default=None)
    subject: Optional[str] = Field(alias="subject", default=None)
    summary: Optional[str] = Field(alias="summary", default=None)
    consequence: Optional[str] = Field(alias="consequence", default=None)
    corrective_action: Optional[str] = Field(alias="correctiveAction", default=None)
    potential_number_of_units_affected: Optional[int] = Field(alias="potentialNumberOfUnitsAffected", default=None)
    notes: Optional[str] = Field(alias="notes", default=None)
    components: List[SafetyIssueComponent] = Field(alias="components", default_factory=list)
    investigations: List[Any] = Field(alias="investigations", default_factory=list) # List of basic investigation info
    associated_documents_count: int = Field(alias="associatedDocumentsCount")
    associated_documents: Union[List[SafetyIssueAssociatedDocument], str] = Field(alias="associatedDocuments", default_factory=list)
    associated_products_count: int = Field(alias="associatedProductsCount")
    associated_products: Union[List[SafetyIssueAssociatedProduct], str] = Field(alias="associatedProducts", default_factory=list)

class SafetyIssueInvestigation(BaseModel):
    nhtsa_action_number: Optional[str] = Field(alias="nhtsaActionNumber", default=None)
    subject: Optional[str] = Field(alias="subject", default=None)
    summary: Optional[str] = Field(alias="summary", default=None)
    date_opened: Optional[datetime] = Field(alias="dateOpened", default=None)
    date_closed: Optional[datetime] = Field(alias="dateClosed", default=None)
    type: Optional[str] = Field(alias="type", default=None)
    components: List[SafetyIssueComponent] = Field(alias="components", default_factory=list)
    associated_documents_count: int = Field(alias="associatedDocumentsCount")
    associated_documents: Union[List[SafetyIssueAssociatedDocument], str] = Field(alias="associatedDocuments", default_factory=list)
    associated_products_count: int = Field(alias="associatedProductsCount")
    associated_products: Union[List[SafetyIssueAssociatedProduct], str] = Field(alias="associatedProducts", default_factory=list)
    recalls: List[SafetyIssueRecallInvestigationInfo] = Field(alias="recalls", default_factory=list) # List of basic recall info

class SafetyIssueManufacturerCommunication(BaseModel):
    manufacturer_communication_number: Optional[str] = Field(alias="manufacturerCommunicationNumber", default=None)
    nhtsa_id_number: int = Field(alias="nhtsaIdNumber")
    subject: Optional[str] = Field(alias="subject", default=None)
    summary: Optional[str] = Field(alias="summary", default=None)
    communication_date: Optional[datetime] = Field(alias="communicationDate", default=None)
    components: List[SafetyIssueComponent] = Field(alias="components", default_factory=list)
    associated_documents_count: int = Field(alias="associatedDocumentsCount")
    associated_documents: Union[List[SafetyIssueAssociatedDocument], str] = Field(alias="associatedDocuments", default_factory=list)
    associated_products_count: int = Field(alias="associatedProductsCount")
    associated_products: Union[List[SafetyIssueAssociatedProduct], str] = Field(alias="associatedProducts", default_factory=list)


class SafetyIssuesDetail(BaseModel):
    complaints: List[SafetyIssueComplaint] = Field(alias="complaints", default_factory=list)
    recalls: List[SafetyIssueRecall] = Field(alias="recalls", default_factory=list)
    investigations: List[SafetyIssueInvestigation] = Field(alias="investigations", default_factory=list)
    manufacturer_communications: List[SafetyIssueManufacturerCommunication] = Field(alias="manufacturerCommunications", default_factory=list)

class SafetyIssueByNhtsaIdResult(BaseModel):
    meta: Meta = Field(alias="meta")
    results: List[SafetyIssuesDetail] = Field(alias="results")

class SafetyIssueGeneralEntry(BaseModel):
    id: int = Field(alias="id")
    artemis_id: Optional[int] = Field(alias="artemisId", default=None)
    campaign_id: Optional[str] = Field(alias="campaignId", default=None)
    consequence: Optional[str] = Field(alias="consequence", default=None)
    corrective_action: Optional[str] = Field(alias="correctiveAction", default=None)
    create_date: Optional[datetime] = Field(alias="createDate", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    issue_year: Optional[str] = Field(alias="issueYear", default=None)
    major: Optional[str] = Field(alias="major", default=None)
    manufacturer_name: Optional[str] = Field(alias="manufacturerName", default=None)
    mfr_recall_number: Optional[str] = Field(alias="mfrRecallNumber", default=None)
    mfr_toll_free_number: Optional[str] = Field(alias="mfrTollFreeNumber", default=None)
    nhtsa_campaign_number: Optional[str] = Field(alias="nhtsaCampaignNumber", default=None)
    notes: Optional[str] = Field(alias="notes", default=None)
    over_the_air_update_yn: Optional[bool] = Field(alias="overTheAirUpdateYn", default=None)
    owner_notification_date: Optional[datetime] = Field(alias="ownerNotificationDate", default=None)
    park_outside_yn: Optional[bool] = Field(alias="parkOutsideYn", default=None)
    park_vehicle_yn: Optional[bool] = Field(alias="parkVehicleYn", default=None)
    potaff: Optional[int] = Field(default=None) # Potential affected
    potdef: Optional[int] = Field(default=None) # Potential defective
    recall_573_date: Optional[datetime] = Field(alias="recall573Date", default=None)
    recall_573_received_date: Optional[datetime] = Field(alias="recall573ReceivedDate", default=None)
    recall_type: Optional[str] = Field(alias="recallType", default=None)
    status_update_date: Optional[datetime] = Field(alias="statusUpdateDate", default=None)
    sub: Optional[str] = Field(default=None)
    subject: Optional[str] = Field(alias="subject", default=None)
    synopsis: Optional[str] = Field(alias="synopsis", default=None)
    type_code: Optional[str] = Field(alias="typeCode", default=None)
    update_date: Optional[datetime] = Field(alias="updateDate", default=None)

class SafetyIssuesListResult(BaseModel):
    meta: Meta = Field(alias="meta")
    results: List[SafetyIssueGeneralEntry] = Field(alias="results")


import httpx
import logging
import asyncio
# import os
import pickle
# import traceback
import time
from typing import Optional, Dict, Any

# Setup logger
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

# Suppress excessive logging from httpx
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.WARNING)


class NhtsaClient:
    """
    The main client for interacting with the NHTSA APIs.
    This client manages HTTP requests, cookies, and rate limiting.
    """
    BASE_URL = "https://api.nhtsa.gov"
    VPIC_BASE_URL = "https://vpic.nhtsa.dot.gov/api"
    STATIC_FILES_BASE_URL = "https://static.nhtsa.gov"
    NRD_BASE_URL = "https://nrd.api.nhtsa.dot.gov"

    def __init__(self, max_concurrent_requests: int = 5, nhtsa_requests_per_minute: int = 100, session_data: Optional[bytes] = None):
        """
        Initializes the NhtsaClient.

        Args:
            max_concurrent_requests (int): The maximum number of concurrent HTTP requests allowed by httpx.
            nhtsa_requests_per_minute (int): The maximum number of requests allowed per minute for the NHTSA API
                                             to respect the server's rate limit.
            session_data (Optional[bytes]): Pickled session data to restore a previous session.
        """
        self.client = httpx.AsyncClient(
            base_url=self.BASE_URL,
            headers={
                "accept": "application/json, text/plain, */*",
                "accept-language": "en-US,en;q=0.9",
                # "user-agent": "NHTSA-SDK/1.0 (Python)",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
            },
            follow_redirects=True,
            timeout=30.0,
            limits=httpx.Limits(max_connections=max_concurrent_requests, max_keepalive_connections=max_concurrent_requests)
        )
        self.vpic_client = httpx.AsyncClient(
            base_url=self.VPIC_BASE_URL,
            headers={
                "accept": "application/json, text/plain, */*",
                "accept-language": "en-US,en;q=0.9",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
            },
            follow_redirects=True,
            timeout=30.0,
            limits=httpx.Limits(max_connections=max_concurrent_requests, max_keepalive_connections=max_concurrent_requests)
        )
        self.static_client = httpx.AsyncClient(
            base_url=self.STATIC_FILES_BASE_URL,
            headers={
                "accept": "*/*",
                "accept-language": "en-US,en;q=0.9",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
            },
            follow_redirects=True,
            timeout=60.0,
            limits=httpx.Limits(max_connections=max_concurrent_requests, max_keepalive_connections=max_concurrent_requests)
        )
        # New client for NRD APIs
        self.nrd_client = httpx.AsyncClient(
            base_url=self.NRD_BASE_URL,
            headers={
                "accept": "application/json, text/plain, */*",
                "accept-language": "en-US,en;q=0.9",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
            },
            follow_redirects=True,
            timeout=30.0,
            limits=httpx.Limits(max_connections=max_concurrent_requests, max_keepalive_connections=max_concurrent_requests)
        )

        self.session_cookies: Dict[str, str] = {}
        
        # Concurrency control for internal requests
        self.semaphore = asyncio.Semaphore(max_concurrent_requests)
        
        # Global Rate limiting for NHTSA server (100 requests/minute)
        self.nhtsa_requests_per_minute = nhtsa_requests_per_minute
        self.nhtsa_request_interval = 60.0 / self.nhtsa_requests_per_minute # Seconds per request (e.g., 0.6 seconds)
        self._last_global_request_completion_time = 0.0 # Tracks when the last request (across all concurrent ones) finished its work and rate limiting sleep
        self._global_rate_limit_lock = asyncio.Lock() # Protects access to _last_global_request_completion_time

        # Initialize API modules
        from .api.safetyservice.index import SafetyServiceAPI
        from .api.recalls.index import RecallsAPI
        from .api.investigations.index import InvestigationsAPI
        from .api.complaints.index import ComplaintsAPI
        from .api.manufacturer_communications.index import ManufacturerCommunicationsAPI
        from .api.car_seat_inspection_locator.index import CarSeatInspectionLocatorAPI
        from .api.vin_decoding.index import VinDecodingAPI
        from .api.static_files.index import StaticFilesAPI
        from .api.vehicle_crash_test_database.index import VehicleCrashTestDatabaseAPI
        from .api.biomechanics_test_database.index import BiomechanicsTestDatabaseAPI
        from .api.component_test_database.index import ComponentTestDatabaseAPI
        from .api.crash_avoidance_test_database.index import CrashAvoidanceTestDatabaseAPI
        from .api.nhtsa_database_code_library.index import NhtsaDatabaseCodeLibraryAPI
        from .api.safety_issues.index import SafetyIssuesAPI
        from .api.products.index import ProductsAPI
        from .api.tag_lookup.index import TagLookupAPI
        from .api.vin_lookup_web.index import VinLookupWebAPI
        from .api.tools.index import ToolsAPI


        self.safety_service = SafetyServiceAPI(self)
        self.recalls = RecallsAPI(self)
        self.investigations = InvestigationsAPI(self)
        self.complaints = ComplaintsAPI(self)
        self.manufacturer_communications = ManufacturerCommunicationsAPI(self)
        self.car_seat_inspection_locator = CarSeatInspectionLocatorAPI(self)
        self.vin_decoding = VinDecodingAPI(self)
        self.static_files = StaticFilesAPI(self)
        self.vehicle_crash_test_database = VehicleCrashTestDatabaseAPI(self)
        self.biomechanics_test_database = BiomechanicsTestDatabaseAPI(self)
        self.component_test_database = ComponentTestDatabaseAPI(self)
        self.crash_avoidance_test_database = CrashAvoidanceTestDatabaseAPI(self)
        self.nhtsa_database_code_library = NhtsaDatabaseCodeLibraryAPI(self)
        self.safety_issues = SafetyIssuesAPI(self)
        self.products = ProductsAPI(self)
        self.tag_lookup = TagLookupAPI(self)
        self.vin_lookup_web = VinLookupWebAPI(self)
        self.tools = ToolsAPI(self)


        # Load from session if provided
        if session_data:
            self._load_from_session_data(session_data)

    def get_session_data(self) -> bytes:
        """
        Serializes the current session state (cookies) into bytes.

        Returns:
            bytes: A pickled dictionary containing the session state.
        """
        try:
            data = {
                "session_cookies": dict(self.client.cookies),
                # Include cookies from other clients if they manage separate cookies
                "vpic_cookies": dict(self.vpic_client.cookies),
                "static_cookies": dict(self.static_client.cookies),
                "nrd_cookies": dict(self.nrd_client.cookies),
            }
            return pickle.dumps(data)
        except Exception as e:
            logger.error(f"Failed to serialize NHTSA session data: {e}", exc_info=True)
            return b''

    def _load_from_session_data(self, session_data: bytes) -> bool:
        """
        Loads a session from a bytes object.

        Args:
            session_data (bytes): The pickled session data.

        Returns:
            bool: True if the session was loaded successfully, False otherwise.
        """
        if not session_data:
            return False
        try:
            data = pickle.loads(session_data)
            self.session_cookies = data.get("session_cookies", {})
            for name, value in self.session_cookies.items():
                self.client.cookies.set(name, value)
                self.vpic_client.cookies.set(name, value)
                self.static_client.cookies.set(name, value)
                self.nrd_client.cookies.set(name, value) # Apply cookies to new NRD client

            # Load specific client cookies if they were stored separately
            for name, value in data.get("vpic_cookies", {}).items():
                self.vpic_client.cookies.set(name, value)
            for name, value in data.get("static_cookies", {}).items():
                self.static_client.cookies.set(name, value)
            for name, value in data.get("nrd_cookies", {}).items():
                self.nrd_client.cookies.set(name, value)


            logger.info("Successfully loaded NHTSA session from provided data.")
            return True
        except Exception as e:
            logger.error(f"Failed to load NHTSA session from data: {e}", exc_info=True)
            return False

    async def _request(self, method: str, path: str, use_vpic_client: bool = False, use_static_client: bool = False, use_nrd_client: bool = False, **kwargs) -> httpx.Response:
        """
        Internal request handler with rate limiting and a retry mechanism for timeouts.

        Args:
            method (str): The HTTP method (e.g., "GET", "POST").
            path (str): The URL path for the request.
            use_vpic_client (bool): If True, use the vPIC client.
            use_static_client (bool): If True, use the static files client.
            use_nrd_client (bool): If True, use the NRD client.
            **kwargs: Additional keyword arguments to pass to httpx.AsyncClient.request.

        Returns:
            httpx.Response: The HTTP response object.

        Raises:
            httpx.RequestError: If an HTTP request fails after multiple retries.
        """
        if use_static_client:
            current_client = self.static_client
        elif use_vpic_client:
            current_client = self.vpic_client
        elif use_nrd_client:
            current_client = self.nrd_client
        else:
            current_client = self.client

        response = None
        for attempt in range(3):  # Try up to 3 times
            try:
                # 1. Acquire the semaphore to limit concurrent HTTP calls to the server
                async with self.semaphore:
                    # 2. Acquire the global rate limit lock to determine if we need to wait
                    async with self._global_rate_limit_lock:
                        now = time.monotonic()
                        # Calculate when the earliest the *next* request (globally) can begin its network operation.
                        next_allowed_time = self._last_global_request_completion_time + self.nhtsa_request_interval
                        sleep_duration = next_allowed_time - now
                        if sleep_duration > 0:
                            await asyncio.sleep(sleep_duration)
                        # Mark the completion time for this request's rate-limiting slot.
                        # This ensures the global rate is maintained, by dictating when the *next* request can start.
                        self._last_global_request_completion_time = time.monotonic()

                    # 3. Make the actual HTTP request (this happens outside the _global_rate_limit_lock,
                    #    allowing concurrent HTTP operations up to max_concurrent_requests once past the rate limit check)
                    response = await current_client.request(method, path, **kwargs)
                    response.raise_for_status() # This will raise for 4xx/5xx responses

                    # Update session cookies after each successful request
                    self.session_cookies.update(response.cookies)
                    for name, value in response.cookies.items():
                        # Propagate cookies to all clients
                        self.client.cookies.set(name, value)
                        self.vpic_client.cookies.set(name, value)
                        self.static_client.cookies.set(name, value)
                        self.nrd_client.cookies.set(name, value)

                    return response

            except httpx.RequestError as e:
                # A request failed (e.g., network error, timeout, or HTTP status error caught by raise_for_status)
                logger.warning(f"Request to {path} failed on attempt {attempt + 1}: {e}. Retrying...", exc_info=True)
                # This failed attempt still consumed a "slot" in terms of rate limiting.
                # The _last_global_request_completion_time was already updated when the request *started* its slot.
                # So, no need to update it again here as it correctly reflects the last point in time a request *began* its slot.
                await asyncio.sleep(2 ** attempt)  # Exponential backoff before next retry
            except Exception as e:
                logger.error(f"An unexpected error occurred during request to {path}: {e}", exc_info=True)
                # For any other exception, re-raise immediately. The rate limit timestamp was set at slot start.
                raise
        # If all retries fail, raise the last encountered error.
        raise httpx.RequestError(f"Failed to complete request to {path} after multiple retries.")

    async def close(self):
        """
        Closes the httpx client sessions.
        """
        await self.client.aclose()
        await self.vpic_client.aclose()
        await self.static_client.aclose()
        await self.nrd_client.aclose() # Close the new NRD client
        logger.info("HTTP client sessions closed.")



from typing import List, Optional, Dict, Any, Union
from pydantic import BaseModel, Field
from datetime import datetime


class BaseNHTSAResponse(BaseModel):
    """Base model for NHTSA API responses."""
    count: Optional[int] = Field(alias="Count")
    message: Optional[str] = Field(alias="Message")
    search_criteria: Optional[str] = Field(alias="SearchCriteria")


class VinDecodeEntry(BaseModel):
    """Represents a single decoded VIN variable in key-value format."""
    value: Optional[str] = Field(alias="Value")
    value_id: Optional[str] = Field(alias="ValueId")
    variable: str = Field(alias="Variable")
    variable_id: int = Field(alias="VariableId")


class VinDecodeResult(BaseNHTSAResponse):
    """Represents the response structure for Decode VIN (key-value)."""
    results: List[VinDecodeEntry] = Field(alias="Results")


class VinDecodeFlatEntry(BaseModel):
    """Represents a single decoded VIN entry in flat format."""
    abs: Optional[str] = Field(alias="ABS", default=None)
    active_safety_sys_note: Optional[str] = Field(alias="ActiveSafetySysNote", default=None)
    adaptive_cruise_control: Optional[str] = Field(alias="AdaptiveCruiseControl", default=None)
    adaptive_driving_beam: Optional[str] = Field(alias="AdaptiveDrivingBeam", default=None)
    adaptive_headlights: Optional[str] = Field(alias="AdaptiveHeadlights", default=None)
    additional_error_text: Optional[str] = Field(alias="AdditionalErrorText", default=None)
    air_bag_loc_curtain: Optional[str] = Field(alias="AirBagLocCurtain", default=None)
    air_bag_loc_front: Optional[str] = Field(alias="AirBagLocFront", default=None)
    air_bag_loc_knee: Optional[str] = Field(alias="AirBagLocKnee", default=None)
    air_bag_loc_seat_cushion: Optional[str] = Field(alias="AirBagLocSeatCushion", default=None)
    air_bag_loc_side: Optional[str] = Field(alias="AirBagLocSide", default=None)
    auto_reverse_system: Optional[str] = Field(alias="AutoReverseSystem", default=None)
    automatic_pedestrian_alerting_sound: Optional[str] = Field(alias="AutomaticPedestrianAlertingSound", default=None)
    axle_configuration: Optional[str] = Field(alias="AxleConfiguration", default=None)
    axles: Optional[str] = Field(alias="Axles", default=None)
    base_price: Optional[str] = Field(alias="BasePrice", default=None)
    battery_a: Optional[str] = Field(alias="BatteryA", default=None)
    battery_a_to: Optional[str] = Field(alias="BatteryA_to", default=None)
    battery_cells: Optional[str] = Field(alias="BatteryCells", default=None)
    battery_info: Optional[str] = Field(alias="BatteryInfo", default=None)
    battery_k_wh: Optional[str] = Field(alias="BatteryKWh", default=None)
    battery_k_wh_to: Optional[str] = Field(alias="BatteryKWh_to", default=None)
    battery_modules: Optional[str] = Field(alias="BatteryModules", default=None)
    battery_packs: Optional[str] = Field(alias="BatteryPacks", default=None)
    battery_type: Optional[str] = Field(alias="BatteryType", default=None)
    battery_v: Optional[str] = Field(alias="BatteryV", default=None)
    battery_v_to: Optional[str] = Field(alias="BatteryV_to", default=None)
    bed_length_in: Optional[str] = Field(alias="BedLengthIN", default=None)
    bed_type: Optional[str] = Field(alias="BedType", default=None)
    blind_spot_intervention: Optional[str] = Field(alias="BlindSpotIntervention", default=None)
    blind_spot_mon: Optional[str] = Field(alias="BlindSpotMon", default=None)
    body_cab_type: Optional[str] = Field(alias="BodyCabType", default=None)
    body_class: Optional[str] = Field(alias="BodyClass", default=None)
    brake_system_desc: Optional[str] = Field(alias="BrakeSystemDesc", default=None)
    brake_system_type: Optional[str] = Field(alias="BrakeSystemType", default=None)
    bus_floor_config_type: Optional[str] = Field(alias="BusFloorConfigType", default=None)
    bus_length: Optional[str] = Field(alias="BusLength", default=None)
    bus_type: Optional[str] = Field(alias="BusType", default=None)
    can_aacn: Optional[str] = Field(alias="CAN_AACN", default=None)
    cib: Optional[str] = Field(alias="CIB", default=None)
    cash_for_clunkers: Optional[str] = Field(alias="CashForClunkers", default=None)
    charger_level: Optional[str] = Field(alias="ChargerLevel", default=None)
    charger_power_kw: Optional[str] = Field(alias="ChargerPowerKW", default=None)
    combined_braking_system: Optional[str] = Field(alias="CombinedBrakingSystem", default=None)
    cooling_type: Optional[str] = Field(alias="CoolingType", default=None)
    curb_weight_lb: Optional[str] = Field(alias="CurbWeightLB", default=None)
    custom_motorcycle_type: Optional[str] = Field(alias="CustomMotorcycleType", default=None)
    daytime_running_light: Optional[str] = Field(alias="DaytimeRunningLight", default=None)
    destination_market: Optional[str] = Field(alias="DestinationMarket", default=None)
    displacement_cc: Optional[str] = Field(alias="DisplacementCC", default=None)
    displacement_ci: Optional[str] = Field(alias="DisplacementCI", default=None)
    displacement_l: Optional[str] = Field(alias="DisplacementL", default=None)
    doors: Optional[str] = Field(alias="Doors", default=None)
    drive_type: Optional[str] = Field(alias="DriveType", default=None)
    driver_assist: Optional[str] = Field(alias="DriverAssist", default=None)
    dynamic_brake_support: Optional[str] = Field(alias="DynamicBrakeSupport", default=None)
    edr: Optional[str] = Field(alias="EDR", default=None)
    esc: Optional[str] = Field(alias="ESC", default=None)
    ev_drive_unit: Optional[str] = Field(alias="EVDriveUnit", default=None)
    electrification_level: Optional[str] = Field(alias="ElectrificationLevel", default=None)
    engine_configuration: Optional[str] = Field(alias="EngineConfiguration", default=None)
    engine_cycles: Optional[str] = Field(alias="EngineCycles", default=None)
    engine_cylinders: Optional[str] = Field(alias="EngineCylinders", default=None)
    engine_hp: Optional[str] = Field(alias="EngineHP", default=None)
    engine_hp_to: Optional[str] = Field(alias="EngineHP_to", default=None)
    engine_kw: Optional[str] = Field(alias="EngineKW", default=None)
    engine_manufacturer: Optional[str] = Field(alias="EngineManufacturer", default=None)
    engine_model: Optional[str] = Field(alias="EngineModel", default=None)
    entertainment_system: Optional[str] = Field(alias="EntertainmentSystem", default=None)
    error_code: Optional[str] = Field(alias="ErrorCode", default=None)
    error_text: Optional[str] = Field(alias="ErrorText", default=None)
    forward_collision_warning: Optional[str] = Field(alias="ForwardCollisionWarning", default=None)
    fuel_injection_type: Optional[str] = Field(alias="FuelInjectionType", default=None)
    fuel_tank_material: Optional[str] = Field(alias="FuelTankMaterial", default=None)
    fuel_tank_type: Optional[str] = Field(alias="FuelTankType", default=None)
    fuel_type_primary: Optional[str] = Field(alias="FuelTypePrimary", default=None)
    fuel_type_secondary: Optional[str] = Field(alias="FuelTypeSecondary", default=None)
    gcwr: Optional[str] = Field(alias="GCWR", default=None)
    gcwr_to: Optional[str] = Field(alias="GCWR_to", default=None)
    gvwr: Optional[str] = Field(alias="GVWR", default=None)
    gvwr_to: Optional[str] = Field(alias="GVWR_to", default=None)
    keyless_ignition: Optional[str] = Field(alias="KeylessIgnition", default=None)
    lane_centering_assistance: Optional[str] = Field(alias="LaneCenteringAssistance", default=None)
    lane_departure_warning: Optional[str] = Field(alias="LaneDepartureWarning", default=None)
    lane_keep_system: Optional[str] = Field(alias="LaneKeepSystem", default=None)
    lower_beam_headlamp_light_source: Optional[str] = Field(alias="LowerBeamHeadlampLightSource", default=None)
    make: Optional[str] = Field(alias="Make", default=None)
    make_id: Optional[str] = Field(alias="MakeID", default=None)
    manufacturer: Optional[str] = Field(alias="Manufacturer", default=None)
    manufacturer_id: Optional[str] = Field(alias="ManufacturerId", default=None)
    model: Optional[str] = Field(alias="Model", default=None)
    model_id: Optional[str] = Field(alias="ModelID", default=None)
    model_year: Optional[str] = Field(alias="ModelYear", default=None)
    motorcycle_chassis_type: Optional[str] = Field(alias="MotorcycleChassisType", default=None)
    motorcycle_suspension_type: Optional[str] = Field(alias="MotorcycleSuspensionType", default=None)
    ncsa_body_type: Optional[str] = Field(alias="NCSABodyType", default=None)
    ncsa_make: Optional[str] = Field(alias="NCSAMake", default=None)
    ncsa_map_exc_approved_by: Optional[str] = Field(alias="NCSAMapExcApprovedBy", default=None)
    ncsa_map_exc_approved_on: Optional[str] = Field(alias="NCSAMapExcApprovedOn", default=None)
    ncsa_mapping_exception: Optional[str] = Field(alias="NCSAMappingException", default=None)
    ncsa_model: Optional[str] = Field(alias="NCSAModel", default=None)
    ncsa_note: Optional[str] = Field(alias="NCSANote", default=None)
    non_land_use: Optional[str] = Field(alias="NonLandUse", default=None)
    note: Optional[str] = Field(alias="Note", default=None)
    other_bus_info: Optional[str] = Field(alias="OtherBusInfo", default=None)
    other_engine_info: Optional[str] = Field(alias="OtherEngineInfo", default=None)
    other_motorcycle_info: Optional[str] = Field(alias="OtherMotorcycleInfo", default=None)
    other_restraint_system_info: Optional[str] = Field(alias="OtherRestraintSystemInfo", default=None)
    other_trailer_info: Optional[str] = Field(alias="OtherTrailerInfo", default=None)
    park_assist: Optional[str] = Field(alias="ParkAssist", default=None)
    pedestrian_automatic_emergency_braking: Optional[str] = Field(alias="PedestrianAutomaticEmergencyBraking", default=None)
    plant_city: Optional[str] = Field(alias="PlantCity", default=None)
    plant_company_name: Optional[str] = Field(alias="PlantCompanyName", default=None)
    plant_country: Optional[str] = Field(alias="PlantCountry", default=None)
    plant_state: Optional[str] = Field(alias="PlantState", default=None)
    possible_values: Optional[str] = Field(alias="PossibleValues", default=None)
    pretensioner: Optional[str] = Field(alias="Pretensioner", default=None)
    rear_automatic_emergency_braking: Optional[str] = Field(alias="RearAutomaticEmergencyBraking", default=None)
    rear_cross_traffic_alert: Optional[str] = Field(alias="RearCrossTrafficAlert", default=None)
    rear_visibility_system: Optional[str] = Field(alias="RearVisibilitySystem", default=None)
    sae_automation_level: Optional[str] = Field(alias="SAEAutomationLevel", default=None)
    sae_automation_level_to: Optional[str] = Field(alias="SAEAutomationLevel_to", default=None)
    seat_belts_all: Optional[str] = Field(alias="SeatBeltsAll", default=None)
    seat_rows: Optional[str] = Field(alias="SeatRows", default=None)
    seats: Optional[str] = Field(alias="Seats", default=None)
    semiautomatic_headlamp_beam_switching: Optional[str] = Field(alias="SemiautomaticHeadlampBeamSwitching", default=None)
    series: Optional[str] = Field(alias="Series", default=None)
    series2: Optional[str] = Field(alias="Series2", default=None)
    steering_location: Optional[str] = Field(alias="SteeringLocation", default=None)
    suggested_vin: Optional[str] = Field(alias="SuggestedVIN", default=None)
    tpms: Optional[str] = Field(alias="TPMS", default=None)
    top_speed_mph: Optional[str] = Field(alias="TopSpeedMPH", default=None)
    track_width: Optional[str] = Field(alias="TrackWidth", default=None)
    traction_control: Optional[str] = Field(alias="TractionControl", default=None)
    trailer_body_type: Optional[str] = Field(alias="TrailerBodyType", default=None)
    trailer_length: Optional[str] = Field(alias="TrailerLength", default=None)
    trailer_type: Optional[str] = Field(alias="TrailerType", default=None)
    transmission_speeds: Optional[str] = Field(alias="TransmissionSpeeds", default=None)
    transmission_style: Optional[str] = Field(alias="TransmissionStyle", default=None)
    trim: Optional[str] = Field(alias="Trim", default=None)
    trim2: Optional[str] = Field(alias="Trim2", default=None)
    turbo: Optional[str] = Field(alias="Turbo", default=None)
    vin: Optional[str] = Field(alias="VIN", default=None)
    valve_train_design: Optional[str] = Field(alias="ValveTrainDesign", default=None)
    vehicle_descriptor: Optional[str] = Field(alias="VehicleDescriptor", default=None)
    vehicle_type: Optional[str] = Field(alias="VehicleType", default=None)
    wheel_base_long: Optional[str] = Field(alias="WheelBaseLong", default=None)
    wheel_base_short: Optional[str] = Field(alias="WheelBaseShort", default=None)
    wheel_base_type: Optional[str] = Field(alias="WheelBaseType", default=None)
    wheel_size_front: Optional[str] = Field(alias="WheelSizeFront", default=None)
    wheel_size_rear: Optional[str] = Field(alias="WheelSizeRear", default=None)
    wheelie_mitigation: Optional[str] = Field(alias="WheelieMitigation", default=None)
    wheels: Optional[str] = Field(alias="Wheels", default=None)
    windows: Optional[str] = Field(alias="Windows", default=None)


class VinDecodeFlatResult(BaseNHTSAResponse):
    """Represents the response structure for Decode VIN (flat format)."""
    results: List[VinDecodeFlatEntry] = Field(alias="Results")


class VinDecodeExtendedResult(BaseNHTSAResponse):
    """Represents the response structure for Decode VIN Extended (key-value)."""
    results: List[VinDecodeEntry] = Field(alias="Results")


class VinDecodeExtendedFlatResult(BaseNHTSAResponse):
    """Represents the response structure for Decode VIN Extended (flat format)."""
    results: List[VinDecodeFlatEntry] = Field(alias="Results")


class WmiDecodeEntry(BaseModel):
    """Represents a single decoded WMI entry."""
    common_name: Optional[str] = Field(alias="CommonName", default=None)
    created_on: Optional[datetime] = Field(alias="CreatedOn")
    date_available_to_public: Optional[datetime] = Field(alias="DateAvailableToPublic")
    make: Optional[str] = Field(alias="Make")
    manufacturer_name: Optional[str] = Field(alias="ManufacturerName")
    parent_company_name: Optional[str] = Field(alias="ParentCompanyName", default=None)
    url: Optional[str] = Field(alias="URL")
    updated_on: Optional[datetime] = Field(alias="UpdatedOn", default=None)
    vehicle_type: Optional[str] = Field(alias="VehicleType")


class WmiDecodeResult(BaseNHTSAResponse):
    """Represents the response structure for Decode WMI."""
    results: List[WmiDecodeEntry] = Field(alias="Results")


class WmiForManufacturerEntry(BaseModel):
    """Represents a single WMI for a manufacturer."""
    country: Optional[str] = Field(alias="Country")
    created_on: Optional[datetime] = Field(alias="CreatedOn")
    date_available_to_public: Optional[datetime] = Field(alias="DateAvailableToPublic")
    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    updated_on: Optional[datetime] = Field(alias="UpdatedOn")
    vehicle_type: str = Field(alias="VehicleType")
    wmi: str = Field(alias="WMI")


class WmiForManufacturerResult(BaseNHTSAResponse):
    """Represents the response structure for Get WMIs for Manufacturer."""
    results: List[WmiForManufacturerEntry] = Field(alias="Results")


class AllMakesEntry(BaseModel):
    """Represents a single make from the Get All Makes endpoint."""
    make_id: int = Field(alias="Make_ID")
    make_name: str = Field(alias="Make_Name")


class AllMakesResult(BaseNHTSAResponse):
    """Represents the response structure for Get All Makes."""
    results: List[AllMakesEntry] = Field(alias="Results")


class PartsEntry(BaseModel):
    """Represents a single part entry from the Get Parts endpoint."""
    cover_letter_url: Optional[str] = Field(alias="CoverLetterURL")
    letter_date: datetime = Field(alias="LetterDate", json_schema_extra={'format': '%m/%d/%Y'})
    manufacturer_id: int = Field(alias="ManufacturerId")
    manufacturer_name: str = Field(alias="ManufacturerName")
    model_year_from: Optional[int] = Field(alias="ModelYearFrom")
    model_year_to: Optional[int] = Field(alias="ModelYearTo")
    name: str = Field(alias="Name")
    type: str = Field(alias="Type")
    url: str = Field(alias="URL")


class PartsResult(BaseNHTSAResponse):
    """Represents the response structure for Get Parts."""
    results: List[PartsEntry] = Field(alias="Results")


class ManufacturerVehicleType(BaseModel):
    """Represents a vehicle type associated with a manufacturer."""
    is_primary: bool = Field(alias="IsPrimary")
    name: str = Field(alias="Name")


class AllManufacturersEntry(BaseModel):
    """Represents a single manufacturer from the Get All Manufacturers endpoint."""
    country: Optional[str] = Field(alias="Country")
    mfr_common_name: Optional[str] = Field(alias="Mfr_CommonName")
    mfr_id: int = Field(alias="Mfr_ID")
    mfr_name: str = Field(alias="Mfr_Name")
    vehicle_types: List[ManufacturerVehicleType] = Field(alias="VehicleTypes")


class AllManufacturersResult(BaseNHTSAResponse):
    """Represents the response structure for Get All Manufacturers."""
    results: List[AllManufacturersEntry] = Field(alias="Results")


class ManufacturerDetailsVehicleType(BaseModel):
    """Represents a detailed vehicle type for a manufacturer."""
    gvwr_from: Optional[str] = Field(alias="GVWRFrom")
    gvwr_to: Optional[str] = Field(alias="GVWRTo")
    is_primary: bool = Field(alias="IsPrimary")
    name: str = Field(alias="Name")


class ManufacturerDetailsEntry(BaseModel):
    """Represents detailed information for a single manufacturer."""
    address: Optional[str] = Field(alias="Address")
    address2: Optional[str] = Field(alias="Address2")
    city: Optional[str] = Field(alias="City")
    contact_email: Optional[str] = Field(alias="ContactEmail")
    contact_fax: Optional[str] = Field(alias="ContactFax")
    contact_phone: Optional[str] = Field(alias="ContactPhone")
    country: Optional[str] = Field(alias="Country")
    db_as: Optional[str] = Field(alias="DBAs")
    equipment_items: List[Dict[str, Any]] = Field(alias="EquipmentItems")  # Could be more specific if schema provided
    last_updated: Optional[datetime] = Field(alias="LastUpdated")
    manufacturer_types: List[Dict[str, str]] = Field(alias="ManufacturerTypes") # Could be more specific
    mfr_common_name: Optional[str] = Field(alias="Mfr_CommonName")
    mfr_id: int = Field(alias="Mfr_ID")
    mfr_name: str = Field(alias="Mfr_Name")
    other_manufacturer_details: Optional[str] = Field(alias="OtherManufacturerDetails")
    postal_code: Optional[str] = Field(alias="PostalCode")
    primary_product: Optional[str] = Field(alias="PrimaryProduct")
    principal_first_name: Optional[str] = Field(alias="PrincipalFirstName")
    principal_last_name: Optional[str] = Field(alias="PrincipalLastName")
    principal_position: Optional[str] = Field(alias="PrincipalPosition")
    state_province: Optional[str] = Field(alias="StateProvince")
    submitted_name: Optional[str] = Field(alias="SubmittedName")
    submitted_on: Optional[datetime] = Field(alias="SubmittedOn")
    submitted_position: Optional[str] = Field(alias="SubmittedPosition")
    vehicle_types: List[ManufacturerDetailsVehicleType] = Field(alias="VehicleTypes")


class ManufacturerDetailsResult(BaseNHTSAResponse):
    """Represents the response structure for Get Manufacturer Details."""
    results: List[ManufacturerDetailsEntry] = Field(alias="Results")


class MakeForManufacturerEntry(BaseModel):
    """Represents a single make for a manufacturer."""
    make_id: int = Field(alias="Make_ID")
    make_name: str = Field(alias="Make_Name")
    mfr_name: str = Field(alias="Mfr_Name")


class MakeForManufacturerResult(BaseNHTSAResponse):
    """Represents the response structure for Get Makes for Manufacturer."""
    results: List[MakeForManufacturerEntry] = Field(alias="Results")


class MakesForManufacturerAndYearEntry(BaseModel):
    """Represents a single make for a manufacturer and year."""
    make_id: int = Field(alias="MakeId")
    make_name: str = Field(alias="MakeName")
    mfr_id: int = Field(alias="MfrId")
    mfr_name: str = Field(alias="MfrName")


class MakesForManufacturerAndYearResult(BaseNHTSAResponse):
    """Represents the response structure for Get Makes for Manufacturer by Manufacturer Name and Year."""
    results: List[MakesForManufacturerAndYearEntry] = Field(alias="Results")


class MakesForVehicleTypeEntry(BaseModel):
    """Represents a single make for a vehicle type."""
    make_id: int = Field(alias="MakeId")
    make_name: str = Field(alias="MakeName")
    vehicle_type_id: int = Field(alias="VehicleTypeId")
    vehicle_type_name: str = Field(alias="VehicleTypeName")


class MakesForVehicleTypeResult(BaseNHTSAResponse):
    """Represents the response structure for Get Makes for Vehicle Type by Vehicle Type Name."""
    results: List[MakesForVehicleTypeEntry] = Field(alias="Results")


class VehicleTypesForMakeEntry(BaseModel):
    """Represents a single vehicle type for a make."""
    make_id: int = Field(alias="MakeId")
    make_name: str = Field(alias="MakeName")
    vehicle_type_id: int = Field(alias="VehicleTypeId")
    vehicle_type_name: str = Field(alias="VehicleTypeName")


class VehicleTypesForMakeResult(BaseNHTSAResponse):
    """Represents the response structure for Get Vehicle Types for Make by Name."""
    results: List[VehicleTypesForMakeEntry] = Field(alias="Results")


class VehicleTypesForMakeIdEntry(BaseModel):
    """Represents a single vehicle type for a make ID."""
    vehicle_type_id: int = Field(alias="VehicleTypeId")
    vehicle_type_name: str = Field(alias="VehicleTypeName")


class VehicleTypesForMakeIdResult(BaseNHTSAResponse):
    """Represents the response structure for Get Vehicle Types for Make by Id."""
    results: List[VehicleTypesForMakeIdEntry] = Field(alias="Results")


class EquipmentPlantCodeEntry(BaseModel):
    """Represents a single equipment plant code."""
    address: Optional[str] = Field(alias="Address", default=None)
    city: Optional[str] = Field(alias="City", default=None)
    country: str = Field(alias="Country")
    dot_code: str = Field(alias="DOTCode")
    name: str = Field(alias="Name")
    old_dot_code: Optional[str] = Field(alias="OldDotCode")
    postal_code: Optional[str] = Field(alias="PostalCode", default=None)
    state_province: Optional[str] = Field(alias="StateProvince", default=None)
    status: Optional[str] = Field(alias="Status", default=None)


class EquipmentPlantCodeResult(BaseNHTSAResponse):
    """Represents the response structure for Get Equipment Plant Codes."""
    results: List[EquipmentPlantCodeEntry] = Field(alias="Results")


class ModelsForMakeEntry(BaseModel):
    """Represents a single model for a make."""
    make_id: int = Field(alias="Make_ID")
    make_name: str = Field(alias="Make_Name")
    model_id: int = Field(alias="Model_ID")
    model_name: str = Field(alias="Model_Name")


class ModelsForMakeResult(BaseNHTSAResponse):
    """Represents the response structure for Get Models for Make."""
    results: List[ModelsForMakeEntry] = Field(alias="Results")


class ModelsForMakeIdEntry(BaseModel):
    """Represents a single model for a make ID."""
    make_id: int = Field(alias="Make_ID")
    make_name: str = Field(alias="Make_Name")
    model_id: int = Field(alias="Model_ID")
    model_name: str = Field(alias="Model_Name")


class ModelsForMakeIdResult(BaseNHTSAResponse):
    """Represents the response structure for Get Models for MakeId."""
    results: List[ModelsForMakeIdEntry] = Field(alias="Results")


class ModelsForMakeYearEntry(BaseModel):
    """Represents a single model for a make, year, and/or vehicle type."""
    make_id: int = Field(alias="Make_ID")
    make_name: str = Field(alias="Make_Name")
    model_id: int = Field(alias="Model_ID")
    model_name: str = Field(alias="Model_Name")
    vehicle_type_id: Optional[int] = Field(alias="VehicleTypeId", default=None)
    vehicle_type_name: Optional[str] = Field(alias="VehicleTypeName", default=None)


class ModelsForMakeYearResult(BaseNHTSAResponse):
    """Represents the response structure for Get Models for Make and a combination of Year and Vehicle Type."""
    results: List[ModelsForMakeYearEntry] = Field(alias="Results")


class VehicleVariableListEntry(BaseModel):
    """Represents a single vehicle variable."""
    data_type: str = Field(alias="DataType")
    description: str = Field(alias="Description")
    group_name: str = Field(alias="GroupName")
    id: int = Field(alias="ID")
    name: str = Field(alias="Name")


class VehicleVariableListResult(BaseNHTSAResponse):
    """Represents the response structure for Get Vehicle Variables List."""
    results: List[VehicleVariableListEntry] = Field(alias="Results")


class VehicleVariableValuesListEntry(BaseModel):
    """Represents a single value for a vehicle variable."""
    element_name: str = Field(alias="ElementName")
    id: int = Field(alias="Id")
    name: str = Field(alias="Name")


class VehicleVariableValuesListResult(BaseNHTSAResponse):
    """Represents the response structure for Get Vehicle Variable Values List."""
    results: List[VehicleVariableValuesListEntry] = Field(alias="Results")


class DecodeVinBatchEntry(VinDecodeFlatEntry):
    """Represents a single VIN decode result from a batch request."""
    # Inherits all fields from VinDecodeFlatEntry


class DecodeVinBatchResult(BaseNHTSAResponse):
    """Represents the response structure for Decode VIN (flat format) in a Batch."""
    results: List[DecodeVinBatchEntry] = Field(alias="Results")


class CanadianVehicleSpecification(BaseModel):
    """Represents a single Canadian vehicle specification."""
    name: str = Field(alias="Name")
    value: str = Field(alias="Value")


class CanadianVehicleSpecificationsEntry(BaseModel):
    """Represents a collection of Canadian vehicle specifications for a model."""
    specs: List[CanadianVehicleSpecification] = Field(alias="Specs")


class CanadianVehicleSpecificationsResult(BaseNHTSAResponse):
    """Represents the response structure for Get Canadian vehicle specifications."""
    results: List[CanadianVehicleSpecificationsEntry] = Field(alias="Results")



from typing import List, Optional, Any
from pydantic import BaseModel, Field


class SafetyRatingMessage(BaseModel):
    """Base model for NHTSA API responses."""
    count: int = Field(alias="Count")
    message: str = Field(alias="Message")


class VehicleVariantResult(BaseModel):
    """Represents a single vehicle variant."""
    vehicle_description: str = Field(alias="VehicleDescription")
    vehicle_id: int = Field(alias="VehicleId")


class VehicleVariant(SafetyRatingMessage):
    """Represents the response for available vehicle variants."""
    results: List[VehicleVariantResult] = Field(alias="Results")


class SafetyRatingDetail(BaseModel):
    """Represents detailed safety rating information for a vehicle."""
    vehicle_picture: Optional[str] = Field(alias="VehiclePicture", default=None)
    overall_rating: Optional[str] = Field(alias="OverallRating", default=None)
    overall_front_crash_rating: Optional[str] = Field(alias="OverallFrontCrashRating", default=None)
    front_crash_driverside_rating: Optional[str] = Field(alias="FrontCrashDriversideRating", default=None)
    front_crash_passengerside_rating: Optional[str] = Field(alias="FrontCrashPassengersideRating", default=None)
    overall_side_crash_rating: Optional[str] = Field(alias="OverallSideCrashRating", default=None)
    side_crash_driverside_rating: Optional[str] = Field(alias="SideCrashDriversideRating", default=None)
    side_crash_passengerside_rating: Optional[str] = Field(alias="SideCrashPassengersideRating", default=None)
    combined_side_barrier_and_pole_rating_front: Optional[str] = Field(alias="combinedSideBarrierAndPoleRating-Front", default=None)
    combined_side_barrier_and_pole_rating_rear: Optional[str] = Field(alias="combinedSideBarrierAndPoleRating-Rear", default=None)
    side_barrier_rating_overall: Optional[str] = Field(alias="sideBarrierRating-Overall", default=None)
    rollover_rating: Optional[str] = Field(alias="RolloverRating", default=None)
    rollover_rating2: Optional[str] = Field(alias="RolloverRating2", default=None)
    rollover_possibility: Optional[float] = Field(alias="RolloverPossibility", default=None)
    rollover_possibility2: Optional[float] = Field(alias="RolloverPossibility2", default=None)
    dynamic_tip_result: Optional[str] = Field(alias="dynamicTipResult", default=None)
    side_pole_crash_rating: Optional[str] = Field(alias="SidePoleCrashRating", default=None)
    nhtsa_electronic_stability_control: Optional[str] = Field(alias="NHTSAElectronicStabilityControl", default=None)
    nhtsa_forward_collision_warning: Optional[str] = Field(alias="NHTSAForwardCollisionWarning", default=None)
    nhtsa_lane_departure_warning: Optional[str] = Field(alias="NHTSALaneDepartureWarning", default=None)
    complaints_count: Optional[int] = Field(alias="ComplaintsCount", default=None)
    recalls_count: Optional[int] = Field(alias="RecallsCount", default=None)
    investigation_count: Optional[int] = Field(alias="InvestigationCount", default=None)
    model_year: Optional[int] = Field(alias="ModelYear", default=None)
    make: Optional[str] = Field(alias="Make", default=None)
    model: Optional[str] = Field(alias="Model", default=None)
    vehicle_description: Optional[str] = Field(alias="VehicleDescription", default=None)
    vehicle_id: Optional[int] = Field(alias="VehicleId", default=None)


class SafetyRatingResult(SafetyRatingMessage):
    """Represents the response for safety ratings for a specific vehicle ID."""
    results: List[SafetyRatingDetail] = Field(alias="Results")


class ModelYearResult(BaseModel):
    """Represents a single model year."""
    model_year: int = Field(alias="ModelYear")
    vehicle_id: int = Field(alias="VehicleId")


class SafetyRatingModelYear(SafetyRatingMessage):
    """Represents the response for all available model years."""
    results: List[ModelYearResult] = Field(alias="Results")


class MakeResult(BaseModel):
    """Represents a single make for a model year."""
    model_year: int = Field(alias="ModelYear")
    make: str = Field(alias="Make")
    vehicle_id: int = Field(alias="VehicleId")


class SafetyRatingMake(SafetyRatingMessage):
    """Represents the response for all makes for a specific model year."""
    results: List[MakeResult] = Field(alias="Results")


class ModelResult(BaseModel):
    """Represents a single model for a make and model year."""
    model_year: int = Field(alias="ModelYear")
    make: str = Field(alias="Make")
    model: str = Field(alias="Model")
    vehicle_id: int = Field(alias="VehicleId")


class SafetyRatingModel(SafetyRatingMessage):
    """Represents the response for all models for a specific make and model year."""
    results: List[ModelResult] = Field(alias="Results")




from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field
from datetime import datetime

from ...lib.models import APIResponse, Error, Meta, Pagination


class NhtsaCaDbTestData(BaseModel):
    """Represents a single crash avoidance test data entry."""
    test_id: Optional[int] = Field(alias="testID", default=None)
    test_submission_test_id: Optional[int] = Field(alias="testSubmissionTestID", default=None)
    modified_date: Optional[str] = Field(alias="modifiedDate", default=None) # Consider datetime if format is consistent
    json_data: Optional[str] = Field(alias="jsonData", default=None) # JSON string, might need further parsing
    comments: Optional[str] = Field(alias="comments", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    modified_by_user_id: Optional[int] = Field(alias="modifiedbyuserid", default=None)
    test_type: Optional[str] = Field(alias="testType", default=None)
    lab_id: Optional[str] = Field(alias="labId", default=None)


class CurveData(BaseModel):
    """Represents curve data for a crash avoidance test."""
    test_id: Optional[int] = Field(alias="testid", default=None)
    curve_no: Optional[int] = Field(alias="curveno", default=None)
    curve_data: Optional[str] = Field(alias="curveData", default=None) # Might be stringified data or a direct curve data string



from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field
from datetime import datetime

# Assuming common models are in lib/models.py
from ...lib.models import APIResponse, Error, Meta, Pagination


class VehicleDocument(BaseModel):
    """Placeholder for vehicle document structure, details not provided in Swagger."""
    # Based on APIResponse results being "array of objects", we'll assume a generic object or more specific if needed
    pass


class VehicleTestData(BaseModel):
    """Placeholder for generic vehicle test data, details not provided in Swagger."""
    pass


class VehicleModel(BaseModel):
    """Placeholder for vehicle model structure, details not provided in Swagger."""
    pass


class OccupantType(BaseModel):
    """Placeholder for occupant type structure, details not provided in Swagger."""
    pass


class VehicleMetadata(BaseModel):
    """Placeholder for vehicle metadata structure, details not provided in Swagger."""
    pass


class VehicleInformation(BaseModel):
    """Placeholder for vehicle information structure, details not provided in Swagger."""
    pass


class VehicleDetailInformation(BaseModel):
    """Placeholder for vehicle detail information structure, details not provided in Swagger."""
    pass


class TestDetail(BaseModel):
    """Placeholder for test detail structure, details not provided in Swagger."""
    pass


class RestraintInformation(BaseModel):
    """Placeholder for restraint information structure, details not provided in Swagger."""
    pass


class OccupantInformation(BaseModel):
    """Placeholder for occupant information structure, details not provided in Swagger."""
    pass


class OccupantDetailInformation(BaseModel):
    """Placeholder for occupant detail information structure, details not provided in Swagger."""
    pass


class MultimediaFile(BaseModel):
    """Placeholder for multimedia file structure, details not provided in Swagger."""
    pass


class IntrusionInformation(BaseModel):
    """Placeholder for intrusion information structure, details not provided in Swagger."""
    pass


class InstrumentationInformation(BaseModel):
    """Placeholder for instrumentation information structure, details not provided in Swagger."""
    pass


class InstrumentationDetailInformation(BaseModel):
    """Placeholder for instrumentation detail information structure, details not provided in Swagger."""
    pass


class BarrierInformation(BaseModel):
    """Placeholder for barrier information structure, details not provided in Swagger."""
    pass


class VehicleInformationResponse(BaseModel):
    """Represents detailed vehicle information for search results."""
    test_no: Optional[int] = Field(alias="testNo", default=None)
    vehicle_no: Optional[int] = Field(alias="vehicleNo", default=None)
    multimedia_files: Optional[str] = Field(alias="multimediaFiles", default=None)
    vehicle_make: Optional[str] = Field(alias="vehicleMake", default=None)
    vehicle_model: Optional[str] = Field(alias="vehicleModel", default=None)
    model_year: Optional[int] = Field(alias="modelYear", default=None)
    engine_type: Optional[str] = Field(alias="engineType", default=None)
    vehicle_test_weight: Optional[str] = Field(alias="vehicleTestWeight", default=None)
    vehicle_length: Optional[str] = Field(alias="vehicleLength", default=None)
    vehicle_speed: Optional[str] = Field(alias="vehicleSpeed", default=None)
    vehicle_width: Optional[str] = Field(alias="vehicleWidth", default=None)
    vax_crush_distance: Optional[str] = Field(alias="vaxCrushDistance", default=None)
    instrumentation_information: Optional[str] = Field(alias="instrumentationInformation", default=None)
    occupant_information: Optional[str] = Field(alias="occupantInformation", default=None)
    barrier_information: Optional[str] = Field(alias="barrierInformation", default=None)


class BarrierInformationResponse(BaseModel):
    """Represents detailed barrier information for search results."""
    test_no: Optional[int] = Field(alias="testNo", default=None)
    multimedia_files: Optional[str] = Field(alias="multimediaFiles", default=None)
    rigid_or_deformable_barrier: Optional[str] = Field(alias="rigidOrDeformableBarrier", default=None)
    barrier_shape: Optional[str] = Field(alias="barrierShape", default=None)
    angle_of_fixed_barrier: Optional[str] = Field(alias="angleofFixedBarrier", default=None)
    diameter_of_pole_barrier: Optional[str] = Field(alias="diameterofPoleBarrier", default=None)
    barrier_commentary: Optional[str] = Field(alias="barrierCommentary", default=None)
    instrumentation_information: Optional[str] = Field(alias="instrumentationInformation", default=None)
    vehicle_information: Optional[str] = Field(alias="vehicleInformation", default=None)



from typing import List, Optional
from pydantic import BaseModel, Field
from datetime import datetime


class ComplaintBaseResponse(BaseModel):
    """Base model for NHTSA API responses."""
    count: int = Field(alias="count")
    message: str = Field(alias="message")


class ProductInfo(BaseModel):
    """Represents product information within a complaint."""
    type: str = Field(alias="type")
    product_year: Optional[str] = Field(alias="productYear")
    product_make: Optional[str] = Field(alias="productMake")
    product_model: Optional[str] = Field(alias="productModel")
    manufacturer: Optional[str] = Field(alias="manufacturer")


class ComplaintResult(BaseModel):
    """Represents a single complaint entry."""
    odi_number: int = Field(alias="odiNumber")
    manufacturer: Optional[str] = Field(alias="manufacturer")
    crash: bool = Field(alias="crash")
    fire: bool = Field(alias="fire")
    number_of_injuries: int = Field(alias="numberOfInjuries")
    number_of_deaths: int = Field(alias="numberOfDeaths")
    date_of_incident: Optional[datetime] = Field(alias="dateOfIncident",
                                               json_schema_extra={'format': '%m/%d/%Y'})
    date_complaint_filed: Optional[datetime] = Field(alias="dateComplaintFiled",
                                                    json_schema_extra={'format': '%m/%d/%Y'})
    vin: Optional[str] = Field(alias="vin")
    components: Optional[str] = Field(alias="components")
    summary: Optional[str] = Field(alias="summary")
    products: List[ProductInfo] = Field(alias="products")


class ComplaintByVehicle(ComplaintBaseResponse):
    """Represents the response structure for complaints by vehicle."""
    results: List[ComplaintResult] = Field(alias="results")


class ModelYearResult(BaseModel):
    """Represents a single model year."""
    model_year: str = Field(alias="modelYear")


class ModelYear(ComplaintBaseResponse):
    """Represents the response structure for all model years for complaints."""
    results: List[ModelYearResult] = Field(alias="results")


class MakeResult(BaseModel):
    """Represents a single make."""
    model_year: str = Field(alias="modelYear")
    make: str = Field(alias="make")


class Make(ComplaintBaseResponse):
    """Represents the response structure for all makes for a model year for complaints."""
    results: List[MakeResult] = Field(alias="results")


class ModelResult(BaseModel):
    """Represents a single model."""
    model_year: str = Field(alias="modelYear")
    make: str = Field(alias="make")
    model: str = Field(alias="model")


class Model(ComplaintBaseResponse):
    """Represents the response structure for all models for a make and model year for complaints."""
    results: List[ModelResult] = Field(alias="results")


class ComplaintByOdiNumber(ComplaintBaseResponse):
    """Represents the response structure for complaints by ODI number."""
    results: List[ComplaintResult] = Field(alias="results")


class ComplaintFlatFile(BaseModel):
    """
    Represents information about an available complaint flat file for download.
    """
    path: str = Field(..., description="The URL path to the complaint flat file.")
    size: str = Field(..., description="The size of the flat file.")
    updated: str = Field(..., description="The last updated date of the flat file.")



from typing import TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from ...client import NhtsaClient

logger = logging.getLogger(__name__)


class StaticFilesAPI:
    """
    API for downloading static files directly from static.nhtsa.gov.
    These are not JSON API endpoints but direct links to documents/archives.
    """
    def __init__(self, client: "NhtsaClient"):
        """
        Initializes the StaticFilesAPI.

        Args:
            client (NhtsaClient): The main client instance.
        """
        self.client = client

    async def download_file(self, file_url: str) -> bytes:
        """
        Downloads a generic static file given its full URL.

        Args:
            file_url (str): The full URL of the static file to download.

        Returns:
            bytes: The content of the downloaded file.

        Raises:
            httpx.RequestError: If the download fails.
        """
        logger.info(f"Downloading static file from: {file_url}")
        response = await self.client._request("GET", file_url, follow_redirects=True, use_static_client=True)
        return response.content

    # Now handled by the undocumented `safety_issues` endpoint
    # async def get_manufacturer_communication_pdf(self, year: int, nhtsa_id_number: int, sequential_number: int = 1) -> bytes:
    #     """
    #     It seems like for the most part, the url to the pdf will be "https://static.nhtsa.gov/odi/tsbs/{PUBLICATION_YEAR}/MC-{NHTSA_DOCUMENT_ID}-0001.pdf"
    #     https://static.nhtsa.gov/odi/tsbs/2022/MC-10213244-0001.pdf
    #     https://static.nhtsa.gov/odi/tsbs/2022/MC-10213315-0001.pdf

    #     However, this seems to not apply to ~10% of bulletins, where instead they have some other number, or an entirely different naming convention. For example:
    #     https://static.nhtsa.gov/odi/tsbs/2016/MC-10213466-9999.pdf
    #     https://static.nhtsa.gov/odi/tsbs/2015/SB-10074303-0699.pdf

    #     """

    #     """
    #     Constructs the URL for a specific Manufacturer Communication PDF and downloads it.

    #     The URL format is: https://static.nhtsa.gov/odi/tsbs/{year}/MC-{nhtsa_id_number}-{sequential_number}.pdf

    #     Args:
    #         year (int): The year the communication was received (from the file path, e.g., 2025).
    #         nhtsa_id_number (int): The NHTSA ID Number (from Field#1 in TSBS.txt, e.g., 11022868).
    #         sequential_number (int): A sequential number for the PDF (e.g., 1 for '0001').

    #     Returns:
    #         bytes: The content of the downloaded PDF file.

    #     Raises:
    #         httpx.RequestError: If the download fails.
    #     """
    #     # Format sequential_number to be 4 digits, e.g., 1 -> 0001
    #     formatted_sequential_number = f"{sequential_number:04d}"
    #     file_url = f"https://static.nhtsa.gov/odi/tsbs/{year}/MC-{nhtsa_id_number}-{formatted_sequential_number}.pdf" # This works for about 70-90% of files.

    #     return await self.download_file(file_url)



from typing import TYPE_CHECKING, List, Optional
from pydantic import parse_obj_as
import logging
from datetime import date

from ...lib.models import APIResponse, Error, Meta, Pagination

if TYPE_CHECKING:
    from ...client import NhtsaClient

logger = logging.getLogger(__name__)


class BiomechanicsTestDatabaseAPI:
    """
    API for programmatic access to the NHTSA Biomechanics Test Database.
    Base URL: /nhtsa/biomechanics/
    """
    def __init__(self, client: "NhtsaClient"):
        """
        Initializes the BiomechanicsTestDatabaseAPI.

        Args:
            client (NhtsaClient): The main client instance.
        """
        self.client = client
        self.base_path = "/nhtsa/biomechanics" # Specific path for this API on the NRD server

    async def find_biomechanics_documents_by_test_no(self, test_no: str) -> APIResponse:
        """
        Retrieves biomechanics documents by test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-documents/test-no/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_all_test_data(self, page_number: Optional[int] = None, order_by: Optional[str] = None, count: int = 20, sort_by: Optional[str] = None) -> APIResponse:
        """
        Retrieves all biomechanics test data.

        Args:
            page_number (Optional[int]): The page number for pagination.
            order_by (Optional[str]): Field to order the results by.
            count (int): Number of rows to return.
            sort_by (Optional[str]): Sort order ("ASC" or "DESC").

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results"
        params = {"count": count}
        if page_number is not None:
            params["pageNumber"] = page_number
        if order_by:
            params["orderBy"] = order_by
        if sort_by:
            params["sortBy"] = sort_by
        response = await self.client._request("GET", url, params=params, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def find_test_data_by_test_reference_no(self, test_reference_no: str) -> APIResponse:
        """
        Retrieves biomechanics test data by test reference number.

        Args:
            test_reference_no (str): The test reference number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/test-reference-no/{test_reference_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def find_test_data_by_test_no(self, test_no: str) -> APIResponse:
        """
        Retrieves biomechanics test data by test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/test-no/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_occupant_types(self) -> APIResponse:
        """
        Retrieves available occupant types for biomechanics tests.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/occupant-types"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_json_export_metadata(self, test_no: str) -> APIResponse:
        """
        Retrieves JSON export metadata for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/metadata/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_test_details(self, test_no: str) -> APIResponse:
        """
        Retrieves test details for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-test-detail/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_restraint_information(self, test_no: str) -> APIResponse:
        """
        Retrieves restraint information for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-restraint-info/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_multimedia_files(self, test_no: str) -> APIResponse:
        """
        Retrieves multimedia files for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-multimedia-files/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_instrumentation_information(self, test_no: str, page_number: Optional[int] = None, order_by: Optional[str] = None, count: int = 20, sort_by: Optional[str] = None) -> APIResponse:
        """
        Retrieves instrumentation information for a given biomechanics test number.

        Args:
            test_no (str): The test number.
            page_number (Optional[int]): The page number for pagination.
            order_by (Optional[str]): Field to order the results by.
            count (int): Number of rows to return.
            sort_by (Optional[str]): Sort order ("ASC" or "DESC").

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-instrumentation-info/{test_no}"
        params = {"count": count}
        if page_number is not None:
            params["pageNumber"] = page_number
        if order_by:
            params["orderBy"] = order_by
        if sort_by:
            params["sortBy"] = sort_by
        response = await self.client._request("GET", url, params=params, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_instrumentation_detail_information(self, curve_no: int, test_no: int) -> APIResponse:
        """
        Retrieves detailed instrumentation information for a specific curve and biomechanics test number.

        Args:
            curve_no (int): The curve number.
            test_no (int): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-instrumentation-detail-info/{curve_no}/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_dummy_occupant_information(self, test_no: str) -> APIResponse:
        """
        Retrieves dummy occupant information for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-dummy-occupant-info/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def get_biological_occupant_information(self, test_no: str) -> APIResponse:
        """
        Retrieves biological occupant information for a given biomechanics test number.

        Args:
            test_no (str): The test number.

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/get-biological-occupant-info/{test_no}"
        response = await self.client._request("GET", url, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())

    async def search_test_data(
        self,
        page_number: Optional[int] = None,
        test_no_from: Optional[int] = None,
        test_no_to: Optional[int] = None,
        test_date_from: Optional[date] = None,
        test_date_to: Optional[date] = None,
        impact_angle_from: Optional[int] = None,
        impact_angle_to: Optional[int] = None,
        closing_speed_from: Optional[float] = None,
        closing_speed_to: Optional[float] = None,
        test_configuration: Optional[str] = None,
        test_ref_number: Optional[str] = None,
        occupant_type: Optional[List[str]] = None,
        contract: Optional[str] = None,
        test_performer: Optional[str] = None,
        order_by: Optional[str] = None,
        count: int = 20,
        sort_by: Optional[str] = None
    ) -> APIResponse:
        """
        Searches for biomechanics test data based on various criteria.

        Args:
            page_number (Optional[int]): The page number for pagination.
            test_no_from (Optional[int]): Test number from.
            test_no_to (Optional[int]): Test number to.
            test_date_from (Optional[date]): Test date from (YYYY-MM-DD).
            test_date_to (Optional[date]): Test date to (YYYY-MM-DD).
            impact_angle_from (Optional[int]): Impact angle from.
            impact_angle_to (Optional[int]): Impact angle to.
            closing_speed_from (Optional[float]): Closing speed from.
            closing_speed_to (Optional[float]): Closing speed to.
            test_configuration (Optional[str]): Test configuration.
            test_ref_number (Optional[str]): Test reference number (supports pattern search).
            occupant_type (Optional[List[str]]): List of occupant types.
            contract (Optional[str]): Contract (supports pattern search).
            test_performer (Optional[str]): Test performer.
            order_by (Optional[str]): Field to order the results by.
            count (int): Number of rows to return.
            sort_by (Optional[str]): Sort order ("ASC" or "DESC").

        Returns:
            APIResponse: A Pydantic model representing the API response.
        """
        url = f"{self.base_path}/api/v1/biomechanics-database-test-results/by-search"
        params = {"count": count}
        if page_number is not None:
            params["pageNumber"] = page_number
        if test_no_from is not None:
            params["testNoFrom"] = test_no_from
        if test_no_to is not None:
            params["testNoTo"] = test_no_to
        if test_date_from:
            params["testDateFrom"] = test_date_from.isoformat()
        if test_date_to:
            params["testDateTo"] = test_date_to.isoformat()
        if impact_angle_from is not None:
            params["impactAngleFrom"] = impact_angle_from
        if impact_angle_to is not None:
            params["impactAngleTo"] = impact_angle_to
        if closing_speed_from is not None:
            params["closingSpeedFrom"] = closing_speed_from
        if closing_speed_to is not None:
            params["closingSpeedTo"] = closing_speed_to
        if test_configuration:
            params["testConfiguration"] = test_configuration
        if test_ref_number:
            params["testRefNumber"] = test_ref_number
        if occupant_type:
            params["occupantType"] = occupant_type
        if contract:
            params["contract"] = contract
        if test_performer:
            params["testPerformer"] = test_performer
        if order_by:
            params["orderBy"] = order_by
        if sort_by:
            params["sortBy"] = sort_by

        response = await self.client._request("GET", url, params=params, use_nrd_client=True)
        return parse_obj_as(APIResponse, response.json())



from typing import TYPE_CHECKING, List, Optional
from pydantic import parse_obj_as
import logging

from .models import RecallByVehicle, ModelYear, Make, Model, RecallCampaign

if TYPE_CHECKING:
    from ...client import NhtsaClient

logger = logging.getLogger(__name__)


class RecallsAPI:
    """
    API for fetching NHTSA Recalls information.
    """
    def __init__(self, client: "NhtsaClient"):
        """
        Initializes the RecallsAPI.

        Args:
            client (NhtsaClient): The main client instance.
        """
        self.client = client

    async def get_recalls_by_vehicle(self, make: str, model: str, model_year: int) -> RecallByVehicle:
        """
        Get recalls for the required combination of Model Year, Make, and Model.

        Args:
            make (str): The make of the vehicle.
            model (str): The model of the vehicle.
            model_year (int): The model year of the vehicle.

        Returns:
            RecallByVehicle: A Pydantic model representing a list of recalls for the given vehicle.
        """
        url = f"/recalls/recallsByVehicle?make={make}&model={model}&modelYear={model_year}"
        response = await self.client._request("GET", url)
        return parse_obj_as(RecallByVehicle, response.json())

    async def get_all_model_years(self) -> ModelYear:
        """
        Request a list of available Model Years for a given product type (Vehicle) for recalls.

        Returns:
            ModelYear: A Pydantic model representing a list of available model years.
        """
        url = "/products/vehicle/modelYears?issueType=r"
        response = await self.client._request("GET", url)
        return parse_obj_as(ModelYear, response.json())

    async def get_all_makes_for_model_year(self, model_year: int) -> Make:
        """
        Request a list of vehicle Makes by providing a specific vehicle Model Year for recalls.

        Args:
            model_year (int): The model year to filter by.

        Returns:
            Make: A Pydantic model representing a list of vehicle makes for that model year.
        """
        url = f"/products/vehicle/makes?modelYear={model_year}&issueType=r"
        response = await self.client._request("GET", url)
        return parse_obj_as(Make, response.json())

    async def get_all_models_for_make_and_model_year(self, model_year: int, make: str) -> Model:
        """
        Request a list of vehicle Models by providing the vehicle Model Year and Make for recalls.

        Args:
            model_year (int): The model year to filter by.
            make (str): The make to filter by.

        Returns:
            Model: A Pydantic model representing a list of models for the given model year and make.
        """
        url = f"/products/vehicle/models?modelYear={model_year}&make={make}&issueType=r"
        response = await self.client._request("GET", url)
        return parse_obj_as(Model, response.json())

    async def get_recalls_by_campaign_number(self, campaign_number: str) -> RecallCampaign:
        """
        Get all recalls as part of a NHTSA recall campaign number.

        Args:
            campaign_number (str): The NHTSA recall campaign number.

        Returns:
            RecallCampaign: A Pydantic model representing a list of recalls for the specified campaign number.
        """
        url = f"/recalls/campaignNumber?campaignNumber={campaign_number}"
        response = await self.client._request("GET", url)
        return parse_obj_as(RecallCampaign, response.json())



from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field
from datetime import datetime

from ...lib.models import APIResponse, Error, Meta, Pagination


class BiomechanicsDocument(BaseModel):
    """Placeholder for biomechanics document structure, details not provided in Swagger."""
    pass


class BiomechanicsTestData(BaseModel):
    """Placeholder for generic biomechanics test data, details not provided in Swagger."""
    pass


class BiomechanicsOccupantType(BaseModel):
    """Placeholder for biomechanics occupant type structure, details not provided in Swagger."""
    pass


class BiomechanicsMetadata(BaseModel):
    """Placeholder for biomechanics metadata structure, details not provided in Swagger."""
    pass


class BiomechanicsTestDetail(BaseModel):
    """Placeholder for biomechanics test detail structure, details not provided in Swagger."""
    pass


class BiomechanicsRestraintInformation(BaseModel):
    """Placeholder for biomechanics restraint information structure, details not provided in Swagger."""
    pass


class BiomechanicsMultimediaFile(BaseModel):
    """Placeholder for biomechanics multimedia file structure, details not provided in Swagger."""
    pass


class BiomechanicsInstrumentationInformation(BaseModel):
    """Placeholder for biomechanics instrumentation information structure, details not provided in Swagger."""
    pass


class BiomechanicsInstrumentationDetailInformation(BaseModel):
    """Placeholder for biomechanics instrumentation detail information structure, details not provided in Swagger."""
    pass


class BiomechanicsDummyOccupantInformation(BaseModel):
    """Placeholder for biomechanics dummy occupant information structure, details not provided in Swagger."""
    pass


class BiomechanicsBiologicalOccupantInformation(BaseModel):
    """Placeholder for biomechanics biological occupant information structure, details not provided in Swagger."""
    pass




from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field
from datetime import datetime

from ...lib.models import APIResponse, Error, Meta, Pagination


class ComponentDocument(BaseModel):
    """Placeholder for component document structure, details not provided in Swagger."""
    pass


class ComponentTestData(BaseModel):
    """Placeholder for generic component test data, details not provided in Swagger."""
    pass


class ComponentOccupantType(BaseModel):
    """Placeholder for component occupant type structure, details not provided in Swagger."""
    pass


class ComponentMetadata(BaseModel):
    """Placeholder for component metadata structure, details not provided in Swagger."""
    pass


class ComponentVehicleInformation(BaseModel):
    """Placeholder for component vehicle information structure, details not provided in Swagger."""
    pass


class ComponentVehicleDetailInformation(BaseModel):
    """Placeholder for component vehicle detail information structure, details not provided in Swagger."""
    pass


class ComponentTestDetail(BaseModel):
    """Placeholder for component test detail structure, details not provided in Swagger."""
    pass


class ComponentMultimediaFile(BaseModel):
    """Placeholder for component multimedia file structure, details not provided in Swagger."""
    pass


class ComponentInstrumentationInformation(BaseModel):
    """Placeholder for component instrumentation information structure, details not provided in Swagger."""
    pass


class ComponentInstrumentationDetailInformation(BaseModel):
    """Placeholder for component instrumentation detail information structure, details not provided in Swagger."""
    pass


class ComponentConfigurationInformation(BaseModel):
    """Placeholder for component configuration information structure, details not provided in Swagger."""
    pass


class ComponentConfigurationDetailInformation(BaseModel):
    """Placeholder for component configuration detail information structure, details not provided in Swagger."""
    pass


class ComponentInformation(BaseModel):
    """Placeholder for component information structure, details not provided in Swagger."""
    pass


class ComponentDetailInformation(BaseModel):
    """Placeholder for component detail information structure, details not provided in Swagger."""
    pass